# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Служебные функции и константы.

import re

CYRRILIC_LOWER_VOWELS = "аоэиуыеёюя"
CYRRILIC_LOWER_CONSONANTS = "йцкнгшщзхъфвпрлджчсмтьб"
VOWELS = "aeiouAEIOUаоэиуыеёюяАОЭИУЫЕЁЮЯ"
CLOSED_SYLLABLE_CHARS = "рлймнРЛЙМН"


def text_to_wordlist(sentence, cyrillic=False):
    regexp = "[^а-яА-Яёa-zA-Z]"
    if cyrillic:
        regexp = "[^а-яА-Яё]"
    sentence = re.sub(regexp, " ", sentence)
    result = sentence.lower().split()
    return result


def text_to_sentences(text):
    regexp = "[\.\?!](?=[\s\n]*[A-ZА-Я])|;|:-|:—|:—|: —|: —|: -"
    regexps = ["(?<=[^A-zА-я][A-ZА-Я])\.", 
                "(?<=[^A-zА-я][A-zА-я])\.[ ]?(?=[A-zА-я][^A-zА-я])",
                "\.(?=,)"
                ]
    for reg in regexps:
        text = "$".join(re.split(reg,text))

    result = re.split(regexp, text)
    result = map(lambda x: x.strip().replace("$", "."), result)
    return result


def to_cyrrilic(text):
    return text.replace("x", "х") \
        .replace("a", "а") \
        .replace("y", "у") \
        .replace("o", "о") \
        .replace("c", "с") \
        .replace("ё", "е")


def normilize_line(text):
    regexp = "[^а-яА-Яёa-zA-Z0-9]"
    text = re.sub(regexp, " ", text)
    result = to_cyrrilic("".join(text.lower().split()))
    return result


def count_vowels(string):
    num_vowels = 0
    for char in string:
        if char in VOWELS:
            num_vowels += 1
    return num_vowels


def get_first_vowel_position(string):
    for i, ch in enumerate(string):
        if ch in VOWELS:
            return i
    return -1


def etree_to_dict(t):
    return {t.tag: map(etree_to_dict, t.iterchildren()) or t.text}


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Класс рифм.

from rupo.generate.prepare.word_form_vocabulary import WordFormVocabulary
from rupo.stress.word import StressedWord
from rupo.util.preprocess import VOWELS


class RhymeProfile:
    def __init__(self, syllable_count: int, stressed_syllable_number: int,
                 stressed_syllable_text: str, next_syllable_text: str, next_char: str):
        self.syllable_count = syllable_count
        self.stressed_syllable_number = stressed_syllable_number
        self.stressed_syllable_text = stressed_syllable_text
        self.next_syllable_text = next_syllable_text
        self.next_char = next_char

    def __str__(self):
        return "Syllable count: {}; Stressed syllable: {}; " \
               "Stressed syllable text: {}; Next syllable: {}; " \
               "Next char: {}".format(self.syllable_count, self.stressed_syllable_number,
                                      self.stressed_syllable_text, self.next_syllable_text, self.next_char)

    def __repr__(self):
        return self.__str__()


class Rhymes(object):
    """
    Поиск рифм.
    """

    @staticmethod
    def is_rhyme(word1: StressedWord, word2: StressedWord, score_border: int=4, syllable_number_border: int=4,
                 word_form_vocabulary: WordFormVocabulary=None) -> bool:
        """
        Проверка рифмованности 2 слов.

        :param word1: первое слово для проверки рифмы, уже акцентуированное (Word).
        :param word2: второе слово для проверки рифмы, уже акцентуированное (Word).
        :param score_border: граница определния рифмы, чем выше, тем строже совпадение.
        :param syllable_number_border: ограничение на номер слога с конца, на который падает ударение.
        :param word_form_vocabulary: словарь словоформ.
        :return result: является рифмой или нет.
        """
        if word_form_vocabulary is not None:
            lemma1 = word_form_vocabulary.get_word_form_by_text(word1.text.lower()).lemma.lower()
            lemma2 = word_form_vocabulary.get_word_form_by_text(word2.text.lower()).lemma.lower()
            if lemma1 == lemma2:
                return False
        profile1 = Rhymes.__get_rhyme_profile(word1)
        profile2 = Rhymes.__get_rhyme_profile(word2)
        score = 0
        for i, ch1 in enumerate(profile1.stressed_syllable_text):
            for j, ch2 in enumerate(profile2.stressed_syllable_text[i:]):
                if ch1 != ch2:
                    continue
                if ch1 in VOWELS:
                    score += 3
                else:
                    score += 1
        if profile1.next_syllable_text == profile2.next_syllable_text and profile1.next_syllable_text != '':
            score += 3
        elif profile1.next_char == profile2.next_char and profile1.next_char != '':
            score += 1
        return (profile1.stressed_syllable_number == profile2.stressed_syllable_number and
                profile1.syllable_count == profile2.syllable_count and
                profile1.stressed_syllable_number <= syllable_number_border and
                score >= score_border)

    @staticmethod
    def __get_rhyme_profile(word: StressedWord) -> 'RhymeProfile':
        """
        Получение профиля рифмовки (набора признаков для сопоставления).

        :param word: уже акцентуированное слово (Word).
        :return profile: профиль рифмовки.
        """
        # TODO: Переход на фонетическое слово, больше признаков.

        profile = RhymeProfile(syllable_count=0,
                               stressed_syllable_number=-1,
                               stressed_syllable_text="",
                               next_syllable_text="",
                               next_char="")
        syllables = list(word.syllables)
        profile.syllable_count = len(syllables)
        for i, syllable in enumerate(reversed(syllables)):
            if syllable.stress == -1:
                continue
            profile.stressed_syllable_text = syllable.text
            profile.stressed_syllable_number = -i-1
            if i != 0:
                profile.next_syllable = syllables[-i].text
            if syllable.stress + 1 < len(word.text):
                profile.next_char = word.text[syllable.stress + 1]
            break
        return profile



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты считывателя разметок.

import unittest

from rupo.files.reader import Reader, FileType
from rupo.stress.predictor import CombinedStressPredictor
from rupo.main.markup import Markup, Line, Word
from rupo.settings import MARKUP_XML_EXAMPLE, TEXT_XML_EXAMPLE, MARKUP_JSON_EXAMPLE


class TestReader(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.stress_predictor = CombinedStressPredictor()

    def test_read(self):
        processed_xml = Reader.read_markups(MARKUP_XML_EXAMPLE, FileType.XML, is_processed=True)
        self.__assert_markup_is_correct(next(processed_xml))

        unprocessed_xml = Reader.read_markups(TEXT_XML_EXAMPLE, FileType.XML, is_processed=False,
                                              stress_predictor=self.stress_predictor)
        self.__assert_markup_is_correct(next(unprocessed_xml))

        processed_json = Reader.read_markups(MARKUP_JSON_EXAMPLE, FileType.JSON, is_processed=True)
        self.__assert_markup_is_correct(next(processed_json))

    def __assert_markup_is_correct(self, markup):
        self.assertIsInstance(markup, Markup)
        self.assertIsNotNone(markup.text)
        self.assertNotEqual(markup.text, "")
        self.assertNotEqual(markup.lines, [])
        self.assertIsInstance(markup.lines[0], Line)
        self.assertIsInstance(markup.lines[0].words[0], Word)



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для токенизатора.

import unittest

from rupo.main.tokenizer import Tokenizer, SentenceTokenizer, Token


class TestTokenizer(unittest.TestCase):
    def test_tokenizer(self):
        text = "О, когда-нибудь, когда?"
        self.assertEqual(Tokenizer.tokenize(text), [
            Token('О', Token.TokenType.WORD, 0, 1),
            Token(',', Token.TokenType.PUNCTUATION, 1, 2),
            Token(' ', Token.TokenType.SPACE, 2, 3),
            Token('когда-нибудь', Token.TokenType.WORD, 3, 15),
            Token(',', Token.TokenType.PUNCTUATION, 15, 16),
            Token(' ', Token.TokenType.SPACE, 16, 17),
            Token('когда', Token.TokenType.WORD, 17, 22),
            Token('?', Token.TokenType.PUNCTUATION, 22, 23)])

        text = " Пора"
        self.assertEqual(Tokenizer.tokenize(text), [
            Token(' ', Token.TokenType.SPACE, 0, 1),
            Token('Пора', Token.TokenType.WORD, 1, 5)])

    def test_numbers(self):
        text = "Очевидно, 1 января 1970 года..."
        self.assertEqual(Tokenizer.tokenize(text), [
            Token('Очевидно', Token.TokenType.WORD, 0, 8),
            Token(',', Token.TokenType.PUNCTUATION, 8, 9),
            Token(' ', Token.TokenType.SPACE, 9, 10),
            Token('1', Token.TokenType.NUMBER, 10, 11),
            Token(' ', Token.TokenType.SPACE, 11, 12),
            Token('января', Token.TokenType.WORD, 12, 18),
            Token(' ', Token.TokenType.SPACE, 18, 19),
            Token('1970', Token.TokenType.NUMBER, 19, 23),
            Token(' ', Token.TokenType.SPACE, 23, 24),
            Token('года', Token.TokenType.WORD, 24, 28),
            Token('...', Token.TokenType.PUNCTUATION, 28, 31)])

        self.assertEqual(Tokenizer.tokenize(text, replace_numbers=True), [
            Token('Очевидно', Token.TokenType.WORD, 0, 8),
            Token(',', Token.TokenType.PUNCTUATION, 8, 9),
            Token(' ', Token.TokenType.SPACE, 9, 10),
            Token('ЧИСЛО', Token.TokenType.WORD, 10, 11),
            Token(' ', Token.TokenType.SPACE, 11, 12),
            Token('января', Token.TokenType.WORD, 12, 18),
            Token(' ', Token.TokenType.SPACE, 18, 19),
            Token('ЧИСЛО', Token.TokenType.WORD, 19, 23),
            Token(' ', Token.TokenType.SPACE, 23, 24),
            Token('года', Token.TokenType.WORD, 24, 28),
            Token('...', Token.TokenType.PUNCTUATION, 28, 31)])


class TestSentenceTokenizer(unittest.TestCase):
    def test_tokenizer(self):
        text1 = "Конкурс учреждён в 2005 году!!! Официальный партнёр конкурса – Президентский центр Б.Н. Ельцина."
        self.assertEqual(SentenceTokenizer.tokenize(text1),
                         ['Конкурс учреждён в 2005 году!!!',
                          'Официальный партнёр конкурса – Президентский центр Б.Н. Ельцина.'])



from pkg_resources import resource_filename
foo_config = resource_filename(__name__, 'foo.conf')

CLASSIFIER_DIR = resource_filename(__name__, "data/classifier/")

DATA_DIR = resource_filename(__name__, "data")

CMU_DICT = resource_filename(__name__, "data/dict/cmu.txt")
ZALYZNYAK_DICT = resource_filename(__name__, "data/dict/zaliznyak.txt")
RU_WIKI_DICT = resource_filename(__name__, "data/dict/wiki_ru.txt")

RU_ALIGNER_DEFAULT_PATH = resource_filename(__name__, "data/g2p_models/ru_aligner.pickle")
EN_ALIGNER_DEFAULT_PATH = resource_filename(__name__, "data/g2p_models/en_aligner.pickle")

RU_GRAPHEME_STRESS_PATH = resource_filename(__name__, "data/dict/ru_grapheme_stress.txt")
RU_GRAPHEME_STRESS_TRIE_PATH = resource_filename(__name__, "data/dict/ru_grapheme_stress.trie")
RU_G2P_DICT_PATH = resource_filename(__name__, "data/dict/ru_g2p.txt")
RU_PHONEME_STRESS_PATH = resource_filename(__name__, "data/dict/ru_phoneme_stress.txt")
RU_PHONEME_STRESS_BIG_PATH = resource_filename(__name__, "data/dict/ru_phoneme_stress_big.txt")
RU_PHONEME_STRESS_TRIE_PATH = resource_filename(__name__, "data/dict/ru_phoneme_stress.trie")
RU_G2P_DEFAULT_MODEL = resource_filename(__name__, "data/g2p_models/g2p_ru_maxlen40_BLSTM256_BLSTM256_dropout0.2_acc992_wer140.h5")
RU_STRESS_DEFAULT_MODEL = resource_filename(__name__, "data/stress_models/stress_ru_LSTM64_dropout0.2_acc99_wer8.h5")

EN_G2P_DICT_PATH = resource_filename(__name__, "data/dict/en_g2p.txt")
EN_PHONEME_STRESS_PATH = resource_filename(__name__, "data/dict/en_phoneme_stress.txt")
EN_PHONEME_STRESS_TRIE_PATH = resource_filename(__name__, "data/dict/en_phoneme_stress.trie")
EN_G2P_DEFAULT_MODEL = resource_filename(__name__, "data/g2p_models/g2p_en_maxlen40_BLSTM256+LSTM256_LSTM128_dropout0.4_acc977_wer379.h5")
EN_STRESS_DEFAULT_MODEL = resource_filename(__name__, "data/stress_models/stress_en_LSTM128_dropout0.2_acc99_wer10.h5")

EXAMPLES_DIR = resource_filename(__name__, "data/examples/")
MARKUP_XML_EXAMPLE = resource_filename(__name__, "data/examples/markup.xml")
MARKUP_JSON_EXAMPLE = resource_filename(__name__, "data/examples/markup.json")
TEXT_XML_EXAMPLE = resource_filename(__name__, "data/examples/text.xml")
TEXT_TXT_EXAMPLE = resource_filename(__name__, "data/examples/text.txt")
HYPHEN_TOKENS = resource_filename(__name__, "data/hyphen-tokens.txt")

G2P_CURRENT_MODEL_DIR = resource_filename(__name__, "data/g2p_models/")
ACCENT_CURRENT_MODEL_DIR = resource_filename(__name__, "data/stress_models/")
GENERATOR_MODEL_DIR = resource_filename(__name__, "data/generator_models/")

GENERATOR_LSTM_MODEL_PATH = resource_filename(__name__, "data/generator_models/lstm.h5")
GENERATOR_MODEL_DESCRIPTION = resource_filename(__name__, "data/generator_models/build.json")
GENERATOR_MODEL_WEIGHTS = resource_filename(__name__, "data/generator_models/weights.h5")
GENERATOR_WORD_FORM_VOCAB_PATH = resource_filename(__name__, "data/generator_models/word_form_vocabulary.pickle")
GENERATOR_VOCAB_PATH = resource_filename(__name__, "data/generator_models/stress_vocabulary.pickle")
GENERATOR_GRAM_VECTORS = resource_filename(__name__, "data/generator_models/grammeme_vectorizer.json")

TEMP_PATH = resource_filename(__name__, "data/temp.txt")

RU_GRAPHEME_SET = " абвгдеёжзийклмнопрстуфхцчшщьыъэюя-"
EN_GRAPHEME_SET = " abcdefghijklmnopqrstuvwxyz.'-"


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для выравнивания g2pц.

import unittest
import logging
import sys

from rupo.g2p.aligner import Aligner
from rupo.settings import RU_G2P_DICT_PATH


class TestAligner(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

    def test_aligner(self):
        aligner = Aligner()
        self.assertEqual(aligner.align('абазия', 'ɐbɐzʲijə'), ('абаз и я', 'ɐbɐzʲijə'))
        self.assertEqual(aligner.align('аахенец', 'aəxʲɪnʲɪʦ'), ('аах ен ец', 'aəxʲɪnʲɪʦ'))
        self.assertEqual(aligner.align('абатский', 'ɐbaʦkʲɪj'), ('абатск ий', 'ɐbaʦ kʲɪj'))
        self.assertEqual(aligner.align('абазинско-русский', 'ɐbɐzʲinskəruskʲɪj'),
                         ('абаз инско-русск ий', 'ɐbɐzʲinskə rus kʲɪj'))
        with open(RU_G2P_DICT_PATH, 'r', encoding='utf-8') as r:
            lines = r.readlines()[:50]
            pairs = [tuple(line.strip().split("\t")) for line in lines]
            for g, p in pairs:
                logging.debug(aligner.align(g, p))




# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Запись в файлы разных расширений.

import os
from typing import List

from rupo.files.reader import RAW_SEPARATOR
from rupo.main.markup import Markup
from rupo.files.reader import FileType


class Writer(object):
    """
    Запись в файл.
    """
    def __init__(self, destination_type: FileType, path: str) -> None:
        """
        Нужно, когда хотим записывать разметки по одной (экономия памяти).

        :param destination_type: тип файла.
        :param path: путь к файлу.
        """
        self.type = destination_type
        self.path = path
        self.file = None
        try:
            os.remove(path)
        except OSError:
            pass

    def open(self) -> None:
        """
        Открываем файл, вызывать до начала записи.
        """
        self.file = open(self.path, "w", encoding="utf-8")
        if self.type == FileType.XML:
            self.file.write('<?xml version="1W.0" encoding="UTF-8"?><items>')

    def write_markup(self, markup: Markup) -> None:
        """
        Запись разметки в уже открытый файл.
        :param markup: разметка.
        """
        assert self.file is not None
        if self.type == FileType.XML:
            xml = markup.to_xml().encode('utf-8')\
                .replace(b'<?xml version="1.0" encoding="UTF-8" ?>', b'').decode('utf-8')
            self.file.write(xml)
        elif self.type == FileType.RAW:
            Writer.__write_markup_raw(markup, self.file)

    def close(self) -> None:
        """
        Закрываем файл.
        """
        if self.type == FileType.XML:
            self.file.write('</items>')
        self.file.close()

    @staticmethod
    def write_markups(destination_type: FileType, markups: List[Markup], path: str) -> None:
        """
        Запись разметок в файл.

        :param destination_type: тип файла.
        :param markups: разметки.
        :param path: путь к файлу.
        """
        with open(path, "w", encoding="utf-8") as file:
            if destination_type == FileType.XML:
                file.write('<?xml version="1.0" encoding="UTF-8"?><items>')
                for markup in markups:
                    xml = markup.to_xml().encode('utf-8')\
                        .replace(b'<?xml version="1.0" encoding="UTF-8" ?>', b'').decode('utf-8')
                    file.write(xml)
                    file.write("\n")
                file.write('</items>')
            elif destination_type == FileType.JSON:
                file.write("[")
                for markup in markups:
                    file.write(markup.to_json())
                    file.write(",")
                file.seek(0, 2)
                size = file.tell()
                file.truncate(size - 1)
                file.write(']')
            elif destination_type == FileType.RAW:
                for markup in markups:
                    Writer.__write_markup_raw(markup, file)

    @staticmethod
    def __write_markup_raw(markup: Markup, file) -> None:
        """
        Запись разметки в текстовом виде (слово+ударение).

        :param markup: разметка.
        :param file: открытый файл.
        """
        lines = []
        for line in markup.lines:
            lines.append(" ".join([word.get_short() for word in line.words]))
        file.write("\n".join(lines))
        file.write(RAW_SEPARATOR)



# -*- coding: utf-8 -*-
# Авторы: Анастасьев Даниил, Гусев Илья
# Описание: Модуль рекуррентой сети для генерации языковой модели.

from itertools import islice
from typing import List, Tuple

import numpy as np
from keras.layers import Input, Embedding, Dense, LSTM, SpatialDropout1D, BatchNormalization, \
    Activation, concatenate, TimeDistributed, Bidirectional
from keras.models import Model, load_model

from rupo.generate.language_model.model_container import ModelContainer
from rupo.generate.language_model.batch_generator import BatchGenerator, CHAR_SET
from rupo.generate.prepare.grammeme_vectorizer import GrammemeVectorizer
from rupo.generate.prepare.loader import CorporaInformationLoader
from rupo.generate.prepare.word_form_vocabulary import WordFormVocabulary
from rupo.settings import GENERATOR_LSTM_MODEL_PATH, GENERATOR_WORD_FORM_VOCAB_PATH, GENERATOR_GRAM_VECTORS


class LSTMGenerator:
    """
    Языковая модель на основе двухуровневой LSTM RNN.
    """
    def __init__(self, embedding_size: int=30000, external_batch_size: int=10000, nn_batch_size: int=768,
                 sentence_maxlen: int=10, lstm_units=368, embeddings_dimension: int=150, 
                 grammeme_dense_units: Tuple[int]=(35, 15), dense_units: int=256, softmax_size: int=60000,
                 dropout: float=0.2, recalculate_softmax=False, max_word_len: int=30, char_embeddings_dimension: int=20,
                 char_lstm_output_dim: int=64):
        """
        :param embedding_size: размер входного слоя (=размер словаря)
        :param softmax_size: размер выхода softmax-слоя (=размер итогового набора вероятностей)
        :param external_batch_size: размер набора семплов для BatchGenerator'а.
        :param nn_batch_size: размер набора семплов для обучения.
        :param sentence_maxlen: маскимальная длина куска предложения.
        """
        self.embedding_size = embedding_size  # type: int
        self.softmax_size = softmax_size  # type: int
        self.external_batch_size = external_batch_size  # type: int
        self.nn_batch_size = nn_batch_size  # type: int
        self.sentence_maxlen = sentence_maxlen  # type: int
        self.word_form_vocabulary = None  # type: WordFormVocabulary
        self.grammeme_vectorizer = None  # type: GrammemeVectorizer
        self.lstm_units = lstm_units  # type: int
        self.embeddings_dimension = embeddings_dimension  # type: int
        self.grammeme_dense_units = grammeme_dense_units  # type: List[int]
        self.dense_units = dense_units  # type: int
        self.dropout = dropout  # type: float
        self.max_word_len = max_word_len  # type: int
        self.char_embeddings_dimension = char_embeddings_dimension  # type: int
        self.char_lstm_output_dim = char_lstm_output_dim  # type: int
        self.model = None  # type: Model
        self.recalculate_softmax = recalculate_softmax  # type: bool

    def prepare(self, filenames: List[str]=list(),
                word_form_vocab_dump_path: str=GENERATOR_WORD_FORM_VOCAB_PATH,
                gram_dump_path: str=GENERATOR_GRAM_VECTORS) -> None:
        """
        Подготовка векторизатора грамматических значений и словаря словоформ по корпусу.
        
        :param filenames: имена файлов с морфоразметкой.
        :param word_form_vocab_dump_path: путь к дампу словаря словоформ.
        :param gram_dump_path: путь к векторам грамматических значений.
        """
        self.grammeme_vectorizer = GrammemeVectorizer(gram_dump_path)
        self.word_form_vocabulary = WordFormVocabulary(word_form_vocab_dump_path)
        if self.grammeme_vectorizer.is_empty() or self.word_form_vocabulary.is_empty():
            loader = CorporaInformationLoader()
            self.word_form_vocabulary, self.grammeme_vectorizer = loader.parse_corpora(filenames)
            self.grammeme_vectorizer.save()
            self.word_form_vocabulary.save()
        if self.recalculate_softmax:
            self.softmax_size = self.word_form_vocabulary.get_softmax_size_by_lemma_size(self.embedding_size)
            print("Recalculated softmax: ", self.softmax_size)

    def save(self, model_filename: str):
        self.model.save(model_filename)

    def load(self, model_filename: str) -> None:
        self.model = load_model(model_filename)

    def build(self):
        """
        Описание модели.
        """
        # Вход лемм
        lemmas = Input(shape=(None,), name='lemmas')
        lemmas_embedding = Embedding(self.embedding_size + 1, self.embeddings_dimension, name='embeddings')(lemmas)
        lemmas_embedding = SpatialDropout1D(.3)(lemmas_embedding)

        # Вход символов
        chars = Input(shape=(None, self.max_word_len), name='chars')
        chars_embedding = Embedding(len(CHAR_SET) + 1, self.char_embeddings_dimension, name='char_embeddings')(chars)
        chars_lstm = TimeDistributed(Bidirectional(
            LSTM(self.char_lstm_output_dim // 2, dropout=self.dropout, recurrent_dropout=self.dropout,
                 return_sequences=False, name='CharLSTM')))(chars_embedding)

        # Вход граммем
        grammemes_input = Input(shape=(None, self.grammeme_vectorizer.grammemes_count()), name='grammemes')
        grammemes_layer = grammemes_input
        for grammeme_dense_layer_units in self.grammeme_dense_units:
            grammemes_layer = Dense(grammeme_dense_layer_units, activation='relu')(grammemes_layer)

        layer = concatenate([lemmas_embedding, grammemes_layer, chars_lstm], name="LSTM_input")
        layer = LSTM(self.lstm_units, dropout=self.dropout, recurrent_dropout=self.dropout,
                     return_sequences=True, name='LSTM_1')(layer)
        layer = LSTM(self.lstm_units, dropout=self.dropout, recurrent_dropout=self.dropout,
                     return_sequences=False, name='LSTM_2')(layer)

        layer = Dense(self.dense_units)(layer)
        layer = BatchNormalization()(layer)
        layer = Activation('relu')(layer)

        output = Dense(self.softmax_size + 1, activation='softmax')(layer)

        self.model = Model(inputs=[lemmas, grammemes_input, chars], outputs=[output])
        self.model.compile(loss='sparse_categorical_crossentropy', optimizer='adam')
        print(self.model.summary())

    @staticmethod
    def __get_validation_data(batch_generator, size):
        """
        Берет первые size батчей из batch_generator для валидационной выборки
        """
        lemmas_list, grammemes_list, chars_list, y_list = [], [], [], []
        for lemmas, grammemes, chars, y in islice(batch_generator, size):
            lemmas_list.append(lemmas)
            grammemes_list.append(grammemes)
            chars_list.append(chars)
            y_list.append(y)
        return np.vstack(lemmas_list), np.vstack(grammemes_list), np.vstack(chars_list), np.hstack(y_list)

    def train(self, filenames: List[str], validation_size: int=5,
              validation_verbosity: int=5, dump_model_freq: int=10,
              save_path: str=GENERATOR_LSTM_MODEL_PATH, start_epoch: int=0) -> None:
        """
        Обучение модели.
        
        :param filenames: имена файлов с морфоразметкой.
        :param validation_size: размер val выборки.
        :param validation_verbosity: каждый validation_verbosity-шаг делается валидация.
        :param dump_model_freq: каждый dump_model_freq-шаг сохраняется модель.
        :param save_path: путь, куда сохранять модель.
        :param start_epoch: эпоха, с которой надо начать.
        """
        batch_generator = BatchGenerator(filenames,
                                         batch_size=self.external_batch_size,
                                         embedding_size=self.embedding_size,
                                         softmax_size=self.softmax_size,
                                         sentence_maxlen=self.sentence_maxlen,
                                         word_form_vocabulary=self.word_form_vocabulary,
                                         grammeme_vectorizer=self.grammeme_vectorizer,
                                         max_word_len=self.max_word_len)

        lemmas_val, grammemes_val, chars_val, y_val = \
            LSTMGenerator.__get_validation_data(batch_generator, validation_size)
        for big_epoch in range(0, 1000):
            print('------------Big Epoch {}------------'.format(big_epoch))
            for epoch, (lemmas, grammemes, chars, y) in enumerate(batch_generator):
                if epoch < start_epoch:
                    continue
                if epoch < validation_size:
                    continue
                self.model.fit([lemmas, grammemes, chars], y, batch_size=self.nn_batch_size, epochs=1, verbose=2)

                if epoch != 0 and epoch % validation_verbosity == 0:
                    print('val loss:', self.model.evaluate([lemmas_val, grammemes_val, chars_val],
                                                           y_val, batch_size=self.nn_batch_size * 2, verbose=0))

                indices = [self.word_form_vocabulary.get_sequence_end_index()]
                for _ in range(10):
                    indices.append(self._sample(self.predict(indices)))
                sentence = [self.word_form_vocabulary.get_word_form_by_index(index) for index in indices]
                print('Sentence', str(big_epoch), str(epoch), end=': ')
                for word in sentence[::-1]:
                    print(word.text, end=' ')
                print()

                if epoch != 0 and epoch % dump_model_freq == 0:
                    self.save(save_path)

    def predict(self, word_indices: List[int]) -> np.array:
        """
        Предсказание вероятностей следующего слова.
        
        :param word_indices: индексы предыдущих слов.
        :return: проекция языковой модели (вероятности следующего слова).
        """
        if len(word_indices) == 0:
            return np.full(self.softmax_size, 1.0 / self.softmax_size, dtype=np.float)

        cur_sent = [self.word_form_vocabulary.get_word_form_by_index(ind) for ind in word_indices]

        x_lemmas = np.zeros((1, len(cur_sent)))
        x_grammemes = np.zeros((1, len(cur_sent), self.grammeme_vectorizer.grammemes_count()))
        x_chars = np.zeros((1, len(cur_sent), self.max_word_len))

        lemmas_vector, grammemes_vector, chars_vector =\
            BatchGenerator.get_sample(cur_sent, self.embedding_size, self.max_word_len,
                                      word_form_vocabulary=self.word_form_vocabulary,
                                      grammeme_vectorizer=self.grammeme_vectorizer)

        x_lemmas[0, -len(cur_sent):] = lemmas_vector
        x_grammemes[0, -len(cur_sent):] = grammemes_vector
        x_chars[0, -len(cur_sent):] = chars_vector
        prob = self.model.predict([x_lemmas, x_grammemes, x_chars], verbose=0)[0]
        return prob

    @staticmethod
    def _sample(prob: np.array, temperature: float=1.0) -> int:
        """
        Выбор слова по набору вероятностей с заданной температурой (распределение Больцмана).
        
        :param prob: вероятности.
        :param temperature: температура.
        :return: индекс итогового слова.
        """
        prob = prob[:-1]  # Для исключения неизвестных слов.
        prob = np.log(prob) / temperature
        prob = np.exp(prob) / np.sum(np.exp(prob))
        return np.random.choice(len(prob), p=prob)


class LSTMModelContainer(ModelContainer):
    """
    Контейнер для языковой модели на основе LSTM.
    """
    def __init__(self, model_path=GENERATOR_LSTM_MODEL_PATH,
                 word_form_vocab_dump_path: str=GENERATOR_WORD_FORM_VOCAB_PATH,
                 gram_dump_path: str=GENERATOR_GRAM_VECTORS, cut_context=10):
        self.lstm = LSTMGenerator(lstm_units=512, dense_units=256, embedding_size=10000, softmax_size=50000)
        self.lstm.prepare(list(), word_form_vocab_dump_path, gram_dump_path)
        self.lstm.load(model_path)
        self.cut_context = cut_context

    def get_model(self, word_indices: List[int]) -> np.array:
        return self.lstm.predict(word_indices[-self.cut_context:])



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для разметки.

import unittest

from rupo.util.data import MARKUP_EXAMPLE
from rupo.main.markup import Markup
from rupo.stress.predictor import CombinedStressPredictor
from rupo.settings import RU_STRESS_DEFAULT_MODEL, ZALYZNYAK_DICT, CMU_DICT, \
    RU_GRAPHEME_STRESS_PATH, RU_GRAPHEME_STRESS_TRIE_PATH


class TestMarkup(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.stress_predictor = CombinedStressPredictor(
            stress_model_path=RU_STRESS_DEFAULT_MODEL,
            zalyzniak_dict=ZALYZNYAK_DICT,
            cmu_dict=CMU_DICT,
            raw_stress_dict_path=RU_GRAPHEME_STRESS_PATH,
            stress_trie_path=RU_GRAPHEME_STRESS_TRIE_PATH
        )
        # cls.stress_predictor = DictStressPredictor(raw_dict_path=RU_GRAPHEME_STRESS_PATH,
        #                                            trie_path=RU_GRAPHEME_STRESS_TRIE_PATH,
        #                                            zalyzniak_dict=ZALYZNYAK_DICT, cmu_dict=CMU_DICT)

    @classmethod
    def tearDownClass(cls):
        del cls.stress_predictor

    def test_from_to(self):
        clean_markup = Markup()
        self.assertEqual(MARKUP_EXAMPLE, clean_markup.from_xml(MARKUP_EXAMPLE.to_xml()))
        clean_markup = Markup()
        self.assertEqual(MARKUP_EXAMPLE, clean_markup.from_json(MARKUP_EXAMPLE.to_json()))

    def test_process_text(self):
        text = "Соломка король себя.\n Пора виться майкой в."
        markup = Markup.process_text(text, self.stress_predictor)
        self.assertEqual(markup, MARKUP_EXAMPLE)




# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты записи разметок.

import unittest
import os

from rupo.main.markup import Markup
from rupo.files.writer import Writer
from rupo.files.reader import Reader, FileType
from rupo.util.data import MARKUP_EXAMPLE
from rupo.settings import EXAMPLES_DIR


class TestWriter(unittest.TestCase):
    def test_write(self):
        temp_file = os.path.join(EXAMPLES_DIR, "temp.xml")
        markup = MARKUP_EXAMPLE
        Writer.write_markups(FileType.XML, [markup], temp_file)
        processed_xml = Reader.read_markups(temp_file, FileType.XML, is_processed=True)
        self.assertEqual(next(processed_xml), markup)
        processed_xml.close()
        os.remove(temp_file)

        temp_file = os.path.join(EXAMPLES_DIR, "temp.txt")
        Writer.write_markups(FileType.RAW, [markup], temp_file)
        processed_raw = Reader.read_markups(temp_file, FileType.RAW, is_processed=True)
        self.assertIsInstance((next(processed_raw)), Markup)
        processed_raw.close()
        os.remove(temp_file)



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Модуль марковских цепей.

import os
import pickle
from collections import Counter, defaultdict
from typing import List, Dict, Tuple

import numpy as np

from rupo.files.reader import Reader, FileType
from rupo.generate.language_model.model_container import ModelContainer
from rupo.main.markup import Markup
from rupo.main.vocabulary import StressVocabulary


class MarkovModelContainer(ModelContainer):
    """
    Марковские цепи.
    """
    def __init__(self, dump_filename: str, vocabulary: StressVocabulary, markup_dump_path: str=None,
                 n_poems: int=None, n_grams: int=2):
        self.n_grams = n_grams
        self.transitions = defaultdict(Counter)  # type: Dict[Tuple, Counter]
        self.vocabulary = vocabulary
        self.dump_filename = dump_filename

        # Делаем дамп модели для ускорения загрузки.
        if os.path.isfile(dump_filename):
            self.load()
        else:
            i = 0
            markups = Reader.read_markups(markup_dump_path, FileType.XML, is_processed=True)
            for markup in markups:
                self.add_markup(markup)
                i += 1
                if n_poems is not None and n_poems == i:
                    break
                if i % 500 == 0:
                    print(i)
            self.save()

    def save(self):
        with open(self.dump_filename, "wb") as f:
            pickle.dump(self.transitions, f, pickle.HIGHEST_PROTOCOL)
        self.vocabulary.save()

    def load(self):
        with open(self.dump_filename, "rb") as f:
            self.transitions = pickle.load(f)
        self.vocabulary.load()

    def generate_chain(self, words: List[int]) -> Dict[Tuple, Counter]:
        """
        Генерация переходов в марковских цепях с учётом частотности.

        :param words: вершины цепи.
        :return: обновленные переходы.
        """
        for i in range(len(words) - self.n_grams + 1):
            current_words = tuple([words[j] for j in range(i, i+self.n_grams-1)])
            next_word = words[i+1]
            self.transitions[current_words][next_word] += 1
        return self.transitions

    def add_markup(self, markup: Markup) -> None:
        """
        Дополнение цепей на основе разметки.

        :param markup: разметка.
        """
        words = []
        for line in markup.lines:
            for word in line.words:
                try:
                    words.append(self.vocabulary.get_word_index(word.to_stressed_word()))
                except IndexError:
                    print("Can't find word in vocabulary.")
                    pass

        # Генерируем переходы.
        self.generate_chain(list(reversed(words)))

    def get_model(self, word_indices: List[int]) -> np.array:
        """
        Получение языковой модели.

        :param word_indices: индексы предыдущих слов.
        :return: языковая модель (распределение вероятностей для следующего слова).
        """
        l = self.vocabulary.size()
        if len(word_indices) < self.n_grams - 1:
            return np.full(self.vocabulary.size(), 1/l, dtype=np.float)
        prev_words = tuple([word_indices[-i] for i in range(1, self.n_grams)])
        if len(self.transitions[prev_words]) == 0:
            return np.full(self.vocabulary.size(), 1/l, dtype=np.float)
        else:
            transition = self.transitions[prev_words]
            s = sum(transition.values())
            model = np.zeros(self.vocabulary.size(), dtype=np.float)
            for index, p in transition.items():
                model[index] = p/s
            return model



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты фонетического разборщика.

import unittest
import logging
import sys

from rupo.g2p.predictor import RNNG2PPredictor
from rupo.settings import RU_G2P_DEFAULT_MODEL


class TestG2Predictor(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.g2p_predictor = RNNG2PPredictor(
            language="ru",
            g2p_model_path=RU_G2P_DEFAULT_MODEL
        )
        logging.basicConfig(stream=sys.stdout, level=logging.DEBUG)

    def test_g2p(self):
        checks = {
            'я': ['ja'],
            'в': ['f'],
            'он': ['on'],
            'корова': ['kərovə', 'kərɐvə'],
            'мышь': ['mɨʂ'],
            'чрезвычайный': ['ʨrʲɪzvɨʨæjnɨj'],
            'абажур': ['əbɐʐur', 'ɐbɐʐur'],
            'лёгкий': ['lʲɵxʲkʲɪj']
        }
        for word, pos in checks.items():
            self.assertIn(self.g2p_predictor.predict(word), pos)



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Рекуррентная сеть для получения фонем из графем.

import numpy as np
import os
from typing import List, Tuple

from sklearn.model_selection import train_test_split
from keras.models import Model, load_model
from keras.callbacks import Callback, EarlyStopping, ModelCheckpoint
from keras.preprocessing import sequence
from keras.layers import LSTM, Bidirectional, Dropout, TimeDistributed, Dense, Input, Embedding
from keras.layers.merge import concatenate
from rupo.settings import RU_GRAPHEME_SET, EN_GRAPHEME_SET
from rupo.g2p.phonemes import Phonemes


class RNNG2PModel:
    phonetic_alphabet = "".join(Phonemes.get_all())

    def __init__(self, dict_path: str=None, word_max_length: int=30, language: str = "ru", rnn=LSTM,
                 units1: int=256, units2: int=256, dropout: float=0.2, batch_size=2048, emb_dimension=30):
        self.rnn = rnn
        self.dropout = dropout  # type: float
        self.units1 = units1  # type: int
        self.units2 = units2  # type: int
        self.language = language  # type: str
        if language == "ru":
            self.grapheme_alphabet = RU_GRAPHEME_SET
        elif language == "en":
            self.grapheme_alphabet = EN_GRAPHEME_SET
        else:
            assert False
        self.dict_path = dict_path  # type: str
        self.word_max_length = word_max_length  # type: int
        self.emb_dimension = emb_dimension
        self.batch_size = batch_size
        self.model = None

    def build(self) -> None:
        """
        Построение модели.
        """
        inp = Input(shape=(None,))

        emb = Embedding(len(self.grapheme_alphabet), self.emb_dimension)(inp)
        encoded = Bidirectional(self.rnn(self.units1, return_sequences=True, recurrent_dropout=self.dropout))(emb)
        encoded = Dropout(self.dropout)(encoded)
        decoded = TimeDistributed(Dense(self.units2, activation="relu"))(encoded)
        predictions = TimeDistributed(Dense(len(self.phonetic_alphabet), activation="softmax"))(decoded)

        model = Model(inputs=inp, outputs=predictions)
        model.compile(loss='sparse_categorical_crossentropy', optimizer='adam', metrics=['accuracy'])
        print(model.summary())
        self.model = model

    def train(self, dir_name: str, enable_checkpoints: bool = False, checkpoint: str = None) -> None:
        """
        Обучение модели.

        :param dir_name: папка с версиями модели.
        :param enable_checkpoints: использовать ли чекпоинты.
        :param checkpoint: загрузка чекпоинта.
        """
        # Подготовка данных
        x, y = self.load_dict()
        x, y = self.prepare_data(x, y)
        # Деление на выборки.
        x_train, x_val, y_train, y_val = train_test_split(x, y, test_size=0.2, random_state=42)
        x_test, x_val, y_test, y_val = train_test_split(x_val, y_val, test_size=0.5, random_state=42)
        # Основные раунды обучения.
        callbacks = [EarlyStopping(monitor='val_acc', patience=2)]  # type: List[Callback]
        if enable_checkpoints:
            checkpoint_name = os.path.join(dir_name, "{epoch:02d}-{val_loss:.2f}.hdf5")
            callbacks.append(ModelCheckpoint(checkpoint_name, monitor='val_loss'))
        if checkpoint is not None:
            self.load(checkpoint)
        self.model.fit(x_train, y_train, verbose=1, epochs=60, batch_size=self.batch_size,
                       validation_data=(x_val, y_val), callbacks=callbacks)
        # Рассчёт точности и word error rate на test выборке.
        accuracy = self.model.evaluate(x_test, y_test)[1]
        wer = self.evaluate_wer(x_test, y_test)[0]
        # Один раунд обучения на всём датасете.
        self.model.fit(x, y, verbose=1, epochs=1, batch_size=self.batch_size)
        # Сохранение модели.
        filename = "g2p_{language}_maxlen{maxlen}_B{rnn}{units1}_B{rnn}{units2}_dropout{dropout}_acc{acc}_wer{wer}.h5"
        filename = filename.format(language=self.language, rnn=self.rnn.__name__,
                                   units1=self.units1, units2=self.units2, dropout=self.dropout,
                                   acc=int(accuracy * 1000), wer=int(wer * 1000), maxlen=self.word_max_length)
        self.model.save(os.path.join(dir_name, filename))

    def predict(self, words: List[str]) -> List[str]:
        """
        Трансляция в фонемы.

        :param words: графические слова.
        :return: фонетические слова.
        """
        x = self.prepare_data(words, None)[0]
        y = self.model.predict(x, verbose=0, batch_size=self.batch_size)
        answers = []
        for word in y:
            answer = ""
            for ch_prob in word:
                i = int(np.argmax(ch_prob))
                answer += self.phonetic_alphabet[i]
            answers.append(answer.strip())
        return answers

    def load(self, filename: str) -> None:
        self.model = load_model(filename)

    def evaluate_wer(self, x: np.array, y: np.array) -> Tuple[float, float]:
        """
        Считаем word error rate - количество слов, в которых была допущена 
        хоть одна ошибка при транскрибировании.

        :param x: данные.
        :param y: ответы
        :return: метрики.
        """
        print("Validation:")
        answer = self.model.predict(x, verbose=0, batch_size=self.batch_size)
        word_errors = 0
        phoneme_errors = 0
        for i, word in enumerate(answer):
            flag = False
            for j, ch_prob in enumerate(word):
                a = np.argmax(ch_prob)
                if y[i][j] != a:
                    phoneme_errors += 1
                    flag = True
            if flag:
                word_errors += 1

        wer = float(word_errors) / answer.shape[0]
        all_phonemes = np.ndarray.flatten(y)
        per = float(phoneme_errors) / len(all_phonemes[all_phonemes != 0])
        print("WER: " + str(wer))
        print("PER(replace only with zeros): " + str(float(phoneme_errors) / len(all_phonemes)))
        print("PER(replace only): " + str(per))
        return wer, per

    def load_dict(self) -> Tuple[List[str], List[str]]:
        """
        Загрузка из словаря g2p.

        :return: графические и фонетические слова.
        """
        x = []
        y = []
        with open(self.dict_path, "r", encoding='utf-8') as f:
            lines = f.readlines()
            for line in lines:
                g = line.split("\t")[0].strip().lower()
                p = line.split("\t")[1].strip()
                flag = False
                for ch in g:
                    if ch not in self.grapheme_alphabet:
                        flag = True
                for ch in p:
                    if ch not in self.phonetic_alphabet:
                        flag = True
                if flag:
                    continue
                x.append(g)
                y.append(p)
        return x, y

    def prepare_data(self, x: List[str], y: List[str] = None) -> Tuple[np.array, np.array]:
        """
        Подготовка данных.

        :param x: графические слова.
        :param y: фонетические слова.
        :return: данные в числовом виде.
        """
        x = [[self.grapheme_alphabet.find(ch) for ch in g] for g in x]
        x = sequence.pad_sequences(x, maxlen=self.word_max_length, value=0)
        if y is not None:
            y = [[self.phonetic_alphabet.find(ch) for ch in p] for p in y]
            y = sequence.pad_sequences(y, maxlen=self.word_max_length, value=0)
            y = y.reshape((y.shape[0], y.shape[1], 1))
        return x, y


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Считыватель файлов разных расширений.

import os
import xml.etree.ElementTree as etree
import json
from enum import Enum
from typing import Iterator

from rupo.main.markup import Markup
from rupo.metre.metre_classifier import MetreClassifier
from rupo.stress.predictor import StressPredictor


RAW_SEPARATOR = "\n\n\n"


class FileType(Enum):
    """
    Тип файла.
    """
    RAW = ".txt"
    XML = ".xml"
    JSON = ".json"
    VOCAB = ".voc"


class Reader(object):
    """
    Считывание из файлов.
    """
    @staticmethod
    def read_markups(path: str, source_type: FileType, is_processed: bool,
                     stress_predictor: StressPredictor=None) -> Iterator[Markup]:
        """
        Считывание разметок (включая разметку по сырым текстам).

        :param path: путь к файлу/папке.
        :param source_type: тип файлов.
        :param is_processed: уже размеченные тексты?
        :param stress_predictor: классификатор ударений (для неразмеченных текстов).
        """
        paths = Reader.get_paths(path, source_type.value)
        for filename in paths:
            with open(filename, "r", encoding="utf-8") as file:
                if is_processed:
                    if source_type == FileType.XML:
                        for elem in Reader.__xml_iter(file, 'markup'):
                            yield Markup().from_xml(etree.tostring(elem, encoding='utf-8', method='xml'))
                    elif source_type == FileType.JSON:
                        j = json.load(file)
                        for item in j['items']:
                            yield Markup().from_dict(item)
                    elif source_type == FileType.RAW:
                        separator_count = 0
                        text = ""
                        for line in file:
                            if line == "\n":
                                separator_count += 1
                            else:
                                text += line
                            if separator_count == 3:
                                separator_count = 0
                                yield Markup().from_raw(text)
                        if text != "":
                            yield Markup().from_raw(text)
                else:
                    assert stress_predictor is not None
                    for text in Reader.read_texts(filename, source_type):
                        yield Reader.__markup_text(text, stress_predictor)

    @staticmethod
    def read_vocabulary(path: str):
        """
        Считывание словаря.

        :param path: путь к словарю.
        :return: слово и его индекс.
        """
        paths = Reader.get_paths(path, FileType.VOCAB.value)
        for filename in paths:
            with open(filename, "r", encoding="utf-8") as file:
                for line in file:
                    fields = line.strip().split('\t')
                    yield Markup().from_raw(fields[0]).lines[0].words[0], int(fields[1])

    @staticmethod
    def read_texts(path: str, source_type: FileType) -> Iterator[str]:
        """
        Считывание текстов.

        :param path: путь к файлу/папке.
        :param source_type: тип файлов.
        """
        paths = Reader.get_paths(path, source_type.value)
        for filename in paths:
            with open(filename, "r", encoding="utf-8") as file:
                if source_type == FileType.XML:
                    for elem in Reader.__xml_iter(file, 'item'):
                        yield elem.find(".//text").text
                elif source_type == FileType.JSON:
                    # TODO: ленивый парсинг
                    j = json.load(file)
                    for item in j['items']:
                        yield item['text']
                elif source_type == FileType.RAW:
                    text = file.read()
                    for t in text.split(RAW_SEPARATOR):
                        yield t

    @staticmethod
    def get_paths(path: str, ext: str) -> Iterator[str]:
        """
        Получение всех файлов заданного типа по заданному пути.

        :param path: путь к файлу/папке.
        :param ext: требуемое расширение.
        """
        if os.path.isfile(path):
            if ext == os.path.splitext(path)[1]:
                yield path
        else:
            for root, folders, files in os.walk(path):
                for file in files:
                    if ext == os.path.splitext(file)[1]:
                        yield os.path.join(root, file)
                for folder in folders:
                    return Reader.get_paths(folder, ext)

    @staticmethod
    def __markup_text(text: str, stress_predictor: StressPredictor) -> Markup:
        """
        Разметка текста.

        :param text: текст.
        :return: разметка.
        """
        markup = Markup.process_text(text, stress_predictor)
        markup = MetreClassifier.improve_markup(markup)[0]
        return markup

    @staticmethod
    def __xml_iter(file, tag):
        """
        :param file: xml файл.
        :param tag: заданный тег.
        :return: все элементы с заданными тегами в xml.
        """
        return (elem for event, elem in etree.iterparse(file, events=['end']) if event == 'end' and elem.tag == tag)



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Класс слова с ударением.

from enum import Enum
from typing import List, Set
from rupo.g2p.graphemes import Graphemes


class Stress:
    """
    Ударение
    """

    class Type(Enum):
        ANY = -1
        PRIMARY = 0
        SECONDARY = 1

    def __init__(self, position: int, stress_type: Type=Type.PRIMARY) -> None:
        self.position = position
        self.type = stress_type

    def __hash__(self):
        return hash(self.position)

    def __eq__(self, other: 'Stress'):
        return self.position == other.position and self.type == other.type

    def __str__(self):
        return str(self.position) + "\t" + str(self.type)

    def __repr__(self):
        return self.__str__()


class StressedWord:
    """
    Слово и его ударения.
    """

    def __init__(self, text: str, stresses: Set[Stress]) -> None:
        self.stresses = stresses
        self.text = text
        self.syllables = Graphemes.get_syllables(text)
        self.__accent_syllables()

    def get_primary_stresses(self) -> List[int]:
        return [stress.position for stress in self.stresses if stress.type == Stress.Type.PRIMARY]

    def get_secondary_stresses(self) -> List[int]:
        return [stress.position for stress in self.stresses if stress.type == Stress.Type.SECONDARY]

    def add_stress(self, position: int, stress_type: Stress.Type=Stress.Type.PRIMARY) -> None:
        self.stresses.add(Stress(position, stress_type))
        self.__accent_syllables()

    def add_stresses(self, stresses: List[Stress]) -> None:
        self.stresses = set(self.stresses).union(set(stresses))
        self.__accent_syllables()

    def __accent_syllables(self):
        for syllable in self.syllables:
            if Stress(syllable.vowel()) in self.stresses:
                syllable.stress = syllable.vowel()
            else:
                syllable.stress = -1

    def __str__(self):
        return self.text + "\t" + ",".join([str(i) for i in self.get_primary_stresses()])+ \
               "\t" + ",".join([str(i) for i in self.get_secondary_stresses()])

    def __repr__(self):
        return self.__str__()

    def __hash__(self):
        return hash(self.text)

    def __eq__(self, other: 'StressedWord'):
        return self.text == other.text



# -*- coding: utf-8 -*-
# Авторы: Гусев Илья, Анастасьев Даниил
# Описание: Модуль векторизатора граммем.

import os
import jsonpickle
from collections import defaultdict
from typing import Dict, List, Set

from rupo.settings import GENERATOR_GRAM_VECTORS
from rupo.util.tqdm_open import tqdm_open


def get_empty_category():
    return {GrammemeVectorizer.UNKNOWN_VALUE}


class GrammemeVectorizer:
    """
    Класс, который собирает возможные грамматические значения по корпусу и на их основе строит грамматические вектора.
    """
    UNKNOWN_VALUE = "Unknown"

    def __init__(self, dump_filename: str=GENERATOR_GRAM_VECTORS):
        """
        :param dump_filename: путь к дампу.
        """
        self.all_grammemes = defaultdict(get_empty_category)  # type: Dict[str, Set]
        self.vectors = []  # type: List[List[int]]
        self.name_to_index = {}  # type: Dict[str, int]
        self.dump_filename = dump_filename  # type: str
        if os.path.exists(self.dump_filename):
            self.load()

    def save(self) -> None:
        with open(self.dump_filename, "w") as f:
            f.write(jsonpickle.encode(self, f))

    def load(self):
        with open(self.dump_filename, "r") as f:
            vectorizer = jsonpickle.decode(f.read())
            self.__dict__.update(vectorizer.__dict__)

    def collect_grammemes(self, filename: str) -> None:
        """
        Собрать возможные грамматические значения по файлу с морфоразметкой.
        
        :param filename: файл с морфоразметкой.
        """
        with tqdm_open(filename, encoding="utf-8") as f:
            for line in f:
                line = line.strip()
                if len(line) == 0:
                    continue
                pos_tag, grammemes = line.split("\t")[2:4]
                self.add_grammemes(pos_tag, grammemes)

    def add_grammemes(self, pos_tag: str, grammemes: str) -> int:
        """
        Добавить новое грамматическое значение в список известных
        """
        grammemes = self.__process_tag(grammemes)
        vector_name = pos_tag + '#' + grammemes
        if vector_name not in self.name_to_index:
            self.name_to_index[vector_name] = len(self.name_to_index)
            self.all_grammemes["POS"].add(pos_tag)
            grammemes = grammemes.split("|") if grammemes != "_" else []
            for grammeme in grammemes:
                category = grammeme.split("=")[0]
                value = grammeme.split("=")[1]
                self.all_grammemes[category].add(value)
        return self.name_to_index[vector_name]

    def init_possible_vectors(self) -> None:
        """
        Инициализировать все возможные векторы по известным грамматическим значениям
        """
        self.vectors = []
        for grammar_val, index in sorted(self.name_to_index.items(), key=lambda x: x[1]):
            pos_tag, grammemes = grammar_val.split('#')
            grammemes = grammemes.split("|") if grammemes != "_" else []
            vector = self.__build_vector(pos_tag, grammemes)
            self.vectors.append(vector)

    def get_vector(self, vector_name: str) -> List[int]:
        """
        Получить вектор по грамматическим значениям.
        
        :param vector_name: часть речи + грамматическое значение.
        :return: вектор.
        """
        if vector_name not in self.name_to_index:
            return [0] * len(self.vectors[0])
        return self.vectors[self.name_to_index[vector_name]]

    def get_vector_by_index(self, index: int) -> List[int]:
        """
        Получить вектор по индексу
        
        :param index: индекс вектора.
        :return: вектор.
        """
        return self.vectors[index] if 0 <= index < len(self.vectors) else [0] * len(self.vectors[0])

    def get_ordered_grammemes(self) -> List[str]:
        """
        Получить упорядоченный список возможных грамматических значений.
        
        :return: список грамматических значений.
        """
        flat = []
        sorted_grammemes = sorted(self.all_grammemes.items(), key=lambda x: x[0])
        for category, values in sorted_grammemes:
            for value in sorted(list(values)):
                flat.append(category+"="+value)
        return flat

    def size(self) -> int:
        return len(self.vectors)

    def grammemes_count(self) -> int:
        return len(self.get_ordered_grammemes())

    def is_empty(self) -> int:
        return len(self.vectors) == 0

    def get_name_by_index(self, index):
        d = {index: name for name, index in self.name_to_index.items()}
        return d[index]

    def get_index_by_name(self, name):
        pos = name.split("#")[0]
        grammemes = self.__process_tag(name.split("#")[1])
        return self.name_to_index[pos + "#" + grammemes]

    def __build_vector(self, pos_tag: str, grammemes: List[str]) -> List[int]:
        """
        Построение вектора по части речи и грамматическим значениям.
        
        :param pos_tag: часть речи.
        :param grammemes: грамматические значения.
        :return: вектор.
        """
        vector = []
        gram_tags = {pair.split("=")[0]: pair.split("=")[1] for pair in grammemes}
        gram_tags["POS"] = pos_tag
        sorted_grammemes = sorted(self.all_grammemes.items(), key=lambda x: x[0])
        for category, values in sorted_grammemes:
            if category not in gram_tags:
                vector += [1 if value == GrammemeVectorizer.UNKNOWN_VALUE else 0 for value in sorted(list(values))]
            else:
                vector += [1 if value == gram_tags[category] else 0 for value in sorted(list(values))]
        return vector

    @staticmethod
    def __process_tag(grammemes):
        return "|".join(sorted(grammemes.strip().split("|")))



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для модуля фонетики.

import unittest

from rupo.main.markup import Syllable
from rupo.util.data import MARKUP_EXAMPLE
from rupo.g2p.graphemes import Graphemes


class TestGraphemes(unittest.TestCase):
    def test_syllables(self):
        checks = {
            'я': [Syllable(0, 1, 0, 'я')],
            'в': [],
            'лдж': [],
            'кронв': [Syllable(0, 5, 0, 'кронв')],
            'он': [Syllable(0, 2, 0, 'он')],
            'когда': [Syllable(0, 2, 0, 'ко'), Syllable(2, 5, 1, 'гда')],
            'майка': [Syllable(0, 3, 0, 'май'), Syllable(3, 5, 1, 'ка')],
            'сонька': [Syllable(0, 4, 0, 'сонь'), Syllable(4, 6, 1, 'ка')],
            'соломка': [Syllable(0, 2, 0, 'со'), Syllable(2, 5, 1, 'лом'), Syllable(5, 7, 2, 'ка')],
            'изжить': [Syllable(0, 1, 0, 'и'), Syllable(1, 6, 1, 'зжить')],
            'виться': [Syllable(0, 2, 0, 'ви'), Syllable(2, 6, 1, 'ться')],
            'данный': [Syllable(0, 2, 0, 'да'), Syllable(2, 6, 1, 'нный')],
            'марка': [Syllable(0, 3, 0, 'мар'), Syllable(3, 5, 1, 'ка')],
            'зорька': [Syllable(0, 4, 0, 'зорь'), Syllable(4, 6, 1, 'ка')],
            'банка': [Syllable(0, 3, 0, 'бан'), Syllable(3, 5, 1, 'ка')],
            'банька': [Syllable(0, 4, 0, 'бань'), Syllable(4, 6, 1, 'ка')],
            'лайка': [Syllable(0, 3, 0, 'лай'), Syllable(3, 5, 1, 'ка')],
            'оттечь': [Syllable(0, 1, 0, 'о'), Syllable(1, 6, 1, 'ттечь')],
            'дяденька': [Syllable(0, 2, 0, 'дя'), Syllable(2, 6, 1, 'день'), Syllable(6, 8, 2, 'ка')],
            'подъезд': [Syllable(0, 2, 0, 'по'), Syllable(2, 7, 1, 'дъезд')],
            'морские': [Syllable(0, 3, 0, 'мор'), Syllable(3, 6, 1, 'ски'), Syllable(6, 7, 2, 'е')],
            'мерзкие': [Syllable(0, 3, 0, 'мер'), Syllable(3, 6, 1, 'зки'), Syllable(6, 7, 2, 'е')],
            'полный': [Syllable(0, 2, 0, 'по'), Syllable(2, 6, 1, 'лный')],
            'зародыш': [Syllable(0, 2, 0, 'за'), Syllable(2, 4, 1, 'ро'), Syllable(4, 7, 2, 'дыш')],
            'война': [Syllable(0, 3, 0, 'вой'), Syllable(3, 5, 1, 'на')],
            'когда-нибудь': [Syllable(0, 2, 0, 'ко'), Syllable(2, 5, 1, 'гда'),
                             Syllable(6, 8, 2, 'ни'), Syllable(8, 12, 3, 'будь')],
        }

        for word, borders in checks.items():
            self.assertEqual(Graphemes.get_syllables(word), borders)



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Класс для определения ударения.

import os
from typing import List
from rupo.stress.phoneme_rnn import RNNPhonemeStressModel
from rupo.stress.grapheme_rnn import RNNGraphemeStressModel
from rupo.stress.dict import StressDict
from rupo.g2p.rnn import RNNG2PModel
from rupo.settings import RU_STRESS_DEFAULT_MODEL, EN_STRESS_DEFAULT_MODEL, RU_G2P_DEFAULT_MODEL, EN_G2P_DEFAULT_MODEL
from rupo.g2p.aligner import Aligner
from rupo.util.preprocess import count_vowels, get_first_vowel_position
from rupo.settings import CMU_DICT, ZALYZNYAK_DICT, RU_GRAPHEME_SET, RU_WIKI_DICT
from rupo.stress.word import Stress


class StressPredictor:
    def predict(self, word: str) -> List[int]:
        raise NotImplementedError()


class RNNGraphemeStressPredictor(StressPredictor):
    def __init__(self, language: str="ru", stress_model_path: str=None):
        self.language = language
        self.stress_model_path = stress_model_path

        if language == "ru":
            self.__init_language_defaults(RU_STRESS_DEFAULT_MODEL)
        elif language == "en":
            self.__init_language_defaults(EN_STRESS_DEFAULT_MODEL)
        else:
            raise RuntimeError("Wrong language")

        if not os.path.exists(self.stress_model_path):
            raise RuntimeError("No stress model available (or wrong path)")

        self.stress_model = RNNGraphemeStressModel(language=language)
        self.stress_model.load(self.stress_model_path)

    def __init_language_defaults(self, stress_model_path):
        if self.stress_model_path is None:
            self.stress_model_path = stress_model_path

    def predict(self, word: str) -> List[int]:
        word = word.lower()
        stresses = self.stress_model.predict([word])[0]
        stresses = [i for i, stress in enumerate(stresses) if stress == 1] + \
                   [i for i, stress in enumerate(stresses) if stress == 2]
        return stresses


class RNNPhonemeStressPredictor(StressPredictor):
    def __init__(self, language: str="ru", stress_model_path: str=None, g2p_model_path: str=None,
                 grapheme_set=RU_GRAPHEME_SET, g2p_dict_path=None, aligner_dump_path=None,
                 ru_wiki_dict=RU_WIKI_DICT, cmu_dict=CMU_DICT):
        self.language = language
        self.stress_model_path = stress_model_path
        self.g2p_model_path = g2p_model_path

        if language == "ru":
            self.__init_language_defaults(RU_STRESS_DEFAULT_MODEL, RU_G2P_DEFAULT_MODEL)
        elif language == "en":
            self.__init_language_defaults(EN_STRESS_DEFAULT_MODEL, EN_G2P_DEFAULT_MODEL)
        else:
            raise RuntimeError("Wrong language")

        if not os.path.exists(self.stress_model_path) or not os.path.exists(self.g2p_model_path):
            raise RuntimeError("No stress or g2p models available (or wrong paths)")

        self.stress_model = RNNPhonemeStressModel(language=language)
        self.stress_model.load(self.stress_model_path)
        self.g2p_model = RNNG2PModel(language=language)
        self.g2p_model.load(self.g2p_model_path)
        self.aligner = Aligner(language, grapheme_set, g2p_dict_path, aligner_dump_path,
                               ru_wiki_dict=ru_wiki_dict, cmu_dict=cmu_dict)

    def __init_language_defaults(self, stress_model_path, g2p_model_path):
        if self.stress_model_path is None:
            self.stress_model_path = stress_model_path
        if self.g2p_model_path is None:
            self.g2p_model_path = g2p_model_path

    def predict(self, word: str) -> List[int]:
        word = word.lower()
        if sum([int(ch not in self.aligner.grapheme_set) for ch in word]) != 0:
            return []
        phonemes = self.g2p_model.predict([word])[0].replace(" ", "")
        stresses = self.stress_model.predict([phonemes])[0]
        stresses = [i for i, stress in enumerate(stresses) if stress == 1 or stress == 2]
        g, p = self.aligner.align(word, phonemes)
        stresses = self.aligner.align_stresses(g, p, stresses, is_grapheme=False)
        for i, stress in enumerate(stresses):
            stresses[i] -= len([ch for ch in g[:stress] if ch == " "])
        stresses = [i for i in stresses if i < len(word)]
        return stresses


class DictStressPredictor(StressPredictor):
    def __init__(self, language="ru", raw_dict_path=None, trie_path=None,
                 zalyzniak_dict=ZALYZNYAK_DICT, cmu_dict=CMU_DICT):
        self.stress_dict = StressDict(language, raw_dict_path=raw_dict_path, trie_path=trie_path,
                                      zalyzniak_dict=zalyzniak_dict, cmu_dict=cmu_dict)

    def predict(self, word: str) -> List[int]:
        """
        Определение ударения в слове по словарю. Возможно несколько вариантов ударения.

        :param word: слово для простановки ударений.
        :return stresses: позиции букв, на которые падает ударение.
        """
        stresses = []
        if count_vowels(word) == 0:
            # Если гласных нет, то и ударений нет.
            pass
        elif count_vowels(word) == 1:
            # Если одна гласная, то на неё и падает ударение.
            stresses.append(get_first_vowel_position(word))
        elif word.find("ё") != -1:
            # Если есть буква "ё", то только на неё может падать ударение.
            stresses.append(word.find("ё"))
        else:
            # Проверяем словарь на наличие форм с ударениями.
            stresses = self.stress_dict.get_stresses(word, Stress.Type.PRIMARY) +\
                       self.stress_dict.get_stresses(word, Stress.Type.SECONDARY)
            if 'е' not in word:
                return stresses
            # Находим все возможные варинаты преобразований 'е' в 'ё'.
            positions = [i for i in range(len(word)) if word[i] == 'е']
            beam = [word[:positions[0]]]
            for i in range(len(positions)):
                new_beam = []
                for prefix in beam:
                    n = positions[i+1] if i+1 < len(positions) else len(word)
                    new_beam.append(prefix + 'ё' + word[positions[i]+1:n])
                    new_beam.append(prefix + 'е' + word[positions[i]+1:n])
                    beam = new_beam
            # И проверяем их по словарю.
            for permutation in beam:
                if len(self.stress_dict.get_stresses(permutation)) != 0:
                    yo_pos = permutation.find("ё")
                    if yo_pos != -1:
                        stresses.append(yo_pos)
        return stresses


class CombinedStressPredictor(StressPredictor):
    def __init__(self, language="ru", stress_model_path: str=None, raw_stress_dict_path=None,
                 stress_trie_path=None, zalyzniak_dict=ZALYZNYAK_DICT, cmu_dict=CMU_DICT):
        self.rnn = RNNGraphemeStressPredictor(language, stress_model_path)
        self.dict = DictStressPredictor(language, raw_stress_dict_path, stress_trie_path, zalyzniak_dict, cmu_dict)

    def predict(self, word: str) -> List[int]:
        stresses = self.dict.predict(word)
        if len(stresses) == 0:
            return self.rnn.predict(word)
        else:
            return stresses



# -*- coding: utf-8 -*-
# Авторы: Гусев Илья, Анастасьев Даниил
# Описание: Словоформа.

from enum import IntEnum


class LemmaCase(IntEnum):
    """
    Тип капитализации словоформы
    """
    NORMAL_CASE = 1  # Может писаться как с большой буквы, так и с маленькой
    PROPER_CASE = 2  # Может писаться только с большой буквы
    UPPER_CASE = 3  # Может писаться только всеми большими


class WordForm(object):
    """
    Класс словоформы.
    """
    def __init__(self, lemma: str, gram_vector_index: int, text: str, case: LemmaCase=LemmaCase.NORMAL_CASE):
        """
        :param lemma: лемма словоформы (=начальная форма, нормальная форма).
        :param gram_vector_index: индекс грамматического вектора.
        :param text: вокабула словоформы.
        """
        self.lemma = lemma  # type: str
        self.gram_vector_index = gram_vector_index  # type: int
        self.text = text  # type: str
        self.case = case
        
    def set_case(self, case: LemmaCase) -> None:
        self.case = case
        
    def get_text_with_case(self) -> str:
        if self.case == LemmaCase.NORMAL_CASE:
            return self.text
        if self.case == LemmaCase.PROPER_CASE:
            return self.text.capitalize()
        if self.case == LemmaCase.UPPER_CASE:
            return self.text.upper()
        
    def __repr__(self):
        return "<Lemma = {}; GrTag = {}; WordForm = {}; LemmaCase = {}>".format(self.lemma,
            self.gram_vector_index, self.text, self.case)

    def __eq__(self, other):
        return (self.lemma, self.gram_vector_index, self.text) == \
               (other.lemma, other.gram_vector_index, other.text)

    def __hash__(self):
        return hash((self.lemma, self.gram_vector_index, self.text))


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Сопоставление шаблону.

from typing import List, Set, Tuple


class TreeNode:
    """
    Нода дерева разбора шаблона.
    """
    leaf_chars = "usUS"
    non_leaf_chars = "*?w"

    def __init__(self, parent: 'TreeNode', children: List['TreeNode'], text: str, pattern_pos: int):
        """
        :param parent: родитель ноды.
        :param children: дети ноды.
        :param text: символ, соответствующий ноде.
        :param pattern_pos: позиция символа в шаблоне
        """
        self.parent = parent  # type: TreeNode
        self.children = children  # type: List[TreeNode]
        self.text = text  # type: str
        self.pattern_pos = pattern_pos  # type: int

    def get_level(self) -> int:
        """
        :return: высота ноды в дереве.
        """
        parent = self.parent
        level = 0
        while parent is not None:
            parent = parent.parent
            level += 1
        return level

    def get_next_sibling(self) -> 'TreeNode':
        """
        :return: соседняя нода справа.
        """
        siblings = self.parent.children
        index = siblings.index(self) + 1
        if index < len(siblings):
            return siblings[index]
        return None

    def get_last_child_leaf(self) -> 'TreeNode':
        """
        :return: последняя нода изе детей, которая является листом.
        """
        for child in reversed(self.children):
            if child.is_leaf():
                return child
        return None

    def is_first_leaf(self) -> bool:
        if not self.is_leaf():
            return False
        return [child for child in self.parent.children if child.is_leaf()][0] == self

    def is_last_leaf(self) -> bool:
        if not self.is_leaf():
            return False
        return [child for child in self.parent.children if child.is_leaf()][-1] == self

    def get_most_left_leaf(self) -> 'TreeNode':
        """
        :return: самый левый потомок.
        """
        node = self
        while len(node.children) != 0:
            node = node.children[0]
        assert node.is_leaf()
        return node

    def print_tree(self) -> None:
        """
        Вывод дерева с корнем в этой ноде.
        """
        stack = list()
        stack.append(self)
        while len(stack) != 0:
            current_node = stack.pop()
            print("\t" * current_node.get_level(), current_node)
            stack += current_node.children

    def is_leaf(self) -> bool:
        """
        :return: является ли нода листом дерева.
        """
        return self.text in TreeNode.leaf_chars

    def __str__(self) -> str:
        return self.text + " " + str(self.pattern_pos)

    def __repr__(self) -> str:
        return self.__str__()

    def __hash__(self):
        return hash(self.pattern_pos)

    def __eq__(self, other):
        return self.pattern_pos == other.pattern_pos


class State:
    """
    Состояние разбора.
    """
    def __init__(self, node: TreeNode, string_pos: int, strong_errors: int, weak_errors: int, pattern: str):
        """
        :param node: нода дерева, соответствующая состоянию.
        :param string_pos: позиция в сопоставляемой строке.
        :param strong_errors: количество ошибок в U и S.
        :param weak_errors: количество ошибок в u и s.
        :param pattern: шаблон - путь, до этого состояния.
        """
        self.node = node  # type: TreeNode
        self.string_pos = string_pos  # type: int
        self.strong_errors = strong_errors  # type: int
        self.weak_errors = weak_errors  # type: int
        self.pattern = pattern  # type: str

    def __str__(self) -> str:
        return str(self.node) + " " + str(self.string_pos) + " " + str(self.strong_errors) + " " + str(self.weak_errors)

    def __repr__(self) -> str:
        return self.__str__()


class PatternAnalyzer:
    """
    Сопоставлятель шаблона и строки.
    """
    def __init__(self, pattern: str, error_border: int=8):
        """
        :param error_border: граница по ошибкам.
        :param pattern: шаблон.
        """
        self.pattern = pattern  # type: str
        self.tree = self.__build_tree(pattern)  # type: TreeNode
        self.error_border = error_border

    @staticmethod
    def count_errors(pattern: str, string: str, error_border: int=8) -> Tuple[str, int, int, bool]:
        """
        :param pattern: шаблон.
        :param string: строка.
        :param error_border: граница по ошибкам.
        :return: лучший шаблон, количество сильных ошибок, количество слабых ошибок.
        """
        analyzer = PatternAnalyzer(pattern, error_border)
        return analyzer.__accept(string)

    @staticmethod
    def __build_tree(pattern: str) -> TreeNode:
        """
        Построение дерева шаблона.

        :param pattern: шаблон.
        :return: корень дерева.
        """
        root_node = TreeNode(None, list(), "R", -1)
        current_node = root_node
        for i, ch in enumerate(pattern):
            if ch == "(":
                node = TreeNode(current_node, list(), "()", i)
                current_node.children.append(node)
                current_node = node
            if ch == ")":
                node = current_node
                current_node = current_node.parent
                # Убираем бессмысленные скобки.
                if i + 1 < len(pattern) and pattern[i + 1] not in "*?":
                    current_node.children = current_node.children[:-1] + node.children
                    for child in node.children:
                        child.parent = current_node
            if ch in TreeNode.leaf_chars:
                current_node.children.append(TreeNode(current_node, list(), ch, i))
            # Заменяем скобки на нетерминалы.
            if ch in TreeNode.non_leaf_chars:
                current_node.children[-1].text = ch
                current_node.children[-1].pattern_pos = i
        return root_node

    def __accept(self, string: str) -> Tuple[str, int, int, bool]:
        """
        :param string: строка.
        :return: лучший шаблон, количество сильных ошибок, количество слабых ошибок, были ли ошибки.
        """
        current_states = [State(None, -1, 0, 0, "")]
        current_node = self.tree.get_most_left_leaf()
        for i, ch in enumerate(string):
            new_states = []
            for state in current_states:
                if state.node is not None:
                    current_node = self.__get_next_leaf(state.node)
                variants = self.__get_variants(current_node)

                # Каждый вариант - новое состояние.
                for variant in variants:
                    assert variant.is_leaf()
                    strong_errors = state.strong_errors + int(variant.text.isupper() and variant.text != ch)
                    weak_errors = state.weak_errors + int(variant.text.islower() and variant.text != ch.lower())
                    new_state = State(variant, i, strong_errors, weak_errors, state.pattern+variant.text)
                    if new_state.strong_errors + new_state.weak_errors > self.error_border:
                        continue
                    new_states.append(new_state)

            if len(new_states) == 0:
                # Можем закончить раньше, если по ошибкам порезали ветки, либо если шаблон меньше строки.
                current_states = PatternAnalyzer.__filter_states(current_states, self.tree)
                pattern, strong_errors, weak_errors = self.__get_min_errors_from_states(current_states)
                diff = (len(string) - i)
                return pattern, strong_errors + diff, weak_errors + diff, True

            current_states = new_states
        current_states = PatternAnalyzer.__filter_states(current_states, self.tree)
        return self.__get_min_errors_from_states(current_states) + (False,)

    @staticmethod
    def __get_variants(current_node: TreeNode) -> Set[TreeNode]:
        """
        :param current_node: текущая нода.
        :return: варианты ноды на том же символе строки, возникают из-за * и ? в шаблоне.
        """
        variants = set()
        current_variant = current_node
        while current_variant is not None:
            if current_variant not in variants:
                variants.add(current_variant)
            else:
                current_variant = current_variant.parent
            current_variant = PatternAnalyzer.__get_next_variant(current_variant)
        return variants

    @staticmethod
    def __get_next_variant(node: TreeNode) -> TreeNode:
        """
        Получение следующего варианта из варинатов текущей ноды.
        
        :param node: текущий вариант.
        :return: следующий вариант. 
        """
        assert node.is_leaf()
        while node.parent is not None:
            parent = node.parent
            grandfather = parent.parent
            uncle = parent.get_next_sibling() if grandfather is not None else None
            is_variable = node.is_first_leaf() or not node.is_leaf()
            if is_variable and uncle is not None:
                return uncle.get_most_left_leaf()
            elif grandfather is not None and grandfather.text == "*" and grandfather.children[-1] == parent:
                return grandfather.get_most_left_leaf()
            if is_variable:
                node = parent
            else:
                break
        return None

    @staticmethod
    def __get_next_leaf(node: TreeNode) -> TreeNode:
        """
        Получение следующей ноды.
        
        :param node: текущая нода.
        :return: следующая нода.
        """
        assert node.is_leaf()
        while node.parent is not None:
            sibling = node.get_next_sibling()
            if sibling is not None:
                return sibling.get_most_left_leaf()
            elif node.parent.text == "*":
                return node.parent.get_most_left_leaf()
            node = node.parent
        return None

    @staticmethod
    def __filter_states(states: List[State], root: TreeNode) -> List[State]:
        """
        Фильтрация по наличию обязательных терминалов.

        :param states: состояния.
        :param root: корень дерева.
        :return: отфильтрованные состояния.
        """
        return [state for state in states if root.get_last_child_leaf() is None or
                state.node.pattern_pos >= root.get_last_child_leaf().pattern_pos]

    @staticmethod
    def __get_min_errors_from_states(states: List[State]) -> Tuple[str, int, int]:
        """
        :param states: состояния.
        :return: лучший шаблон, количество сильных ошибок, количество слабых ошибок.
        """
        if len(states) == 0:
            return "", 0, 0
        return min([(state.pattern, state.strong_errors, state.weak_errors) for i, state in enumerate(states)],
                   key=lambda x: (x[1], x[2], x[0]))



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Индексы слов для языковой модели.

from typing import Dict
import pickle
import os

from rupo.main.markup import Markup
from rupo.files.reader import Reader, FileType
from rupo.stress.word import StressedWord


class StressVocabulary(object):
    """
    Индексированный словарь.
    """
    def __init__(self, dump_filename: str, markup_path: str=None, from_voc: bool=False) -> None:
        """
        :param dump_filename: файл, в который сохранется словарь.
        :param markup_path: файл/папка с разметками.
        """
        self.dump_filename = dump_filename
        self.word_to_index = {}  # type: Dict[StressedWord, int]
        self.index_to_word = {}  # type: Dict[int, StressedWord]

        if os.path.isfile(self.dump_filename):
            self.load()
        elif markup_path is not None:
            if from_voc:
                word_indexes = Reader.read_vocabulary(markup_path)
                for word, index in word_indexes:
                    self.add_word(word.to_stressed_word(), index)
            else:
                markups = Reader.read_markups(markup_path, FileType.XML, is_processed=True)
                for markup in markups:
                    self.add_markup(markup)
            self.save()

    def save(self) -> None:
        """
        Сохранение словаря.
        """
        with open(self.dump_filename, "wb") as f:
            pickle.dump(self, f, pickle.HIGHEST_PROTOCOL)

    def load(self):
        """
        Загрузка словаря.
        """
        with open(self.dump_filename, "rb") as f:
            vocab = pickle.load(f)
            self.__dict__.update(vocab.__dict__)

    def add_markup(self, markup: Markup) -> None:
        """
        Добавление слов из разметки в словарь.

        :param markup: разметка.
        """
        for line in markup.lines:
            for word in line.words:
                self.add_word(word.to_stressed_word())

    def add_word(self, word: StressedWord, index: int=-1) -> bool:
        """
        Добавление слова.

        :param word: слово.
        :param index: индекс, если задан заранее.
        :return: слово новое или нет.
        """
        if word in self.word_to_index:
            if index != -1:
                self.index_to_word[index] = word
            return False
        self.word_to_index[word] = self.size() if index == -1 else index
        self.index_to_word[self.size() if index == -1 else index] = word
        return True

    def get_word_index(self, word: StressedWord) -> int:
        """
        Получить индекс слова.

        :param word: слово (Word).
        :return: индекс.
        """
        if word in self.word_to_index:
            return self.word_to_index[word]
        raise IndexError("Can't find word: " + word.text)

    def get_word(self, index: int) -> StressedWord:
        """
        Получить слово по индексу.

        :param index: индекс.
        :return: слово.
        """
        return self.index_to_word[index]

    def size(self):
        """
        :return: получить размер словаря.
        """
        return len(self.index_to_word)



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Модуль токенизации.

import re
from typing import List
from enum import Enum, unique

from rupo.settings import HYPHEN_TOKENS


class Token:
    @unique
    class TokenType(Enum):
        """
        Тип токена.
        """
        UNKNOWN = -1
        WORD = 0
        PUNCTUATION = 1
        SPACE = 2
        ENDLINE = 3
        NUMBER = 4

        def __str__(self):
            return str(self.name)

        def __repr__(self):
            return self.__str__()

    def __init__(self, text: str, token_type: TokenType, begin: int, end: int):
        """
        :param text: исходный текст.
        :param token_type: тип токена.
        :param begin: начало позиции токена в тексте.
        :param end: конец позиции токена в тексте.
        """
        self.token_type = token_type
        self.begin = begin
        self.end = end
        self.text = text

    def __str__(self):
        return "'" + self.text + "'" + "|" + str(self.token_type) + " (" + str(self.begin) + ", " + str(self.end) + ")"

    def __repr__(self):
        return self.__str__()

    def __eq__(self, other):
        return self.text == other.text and self.token_type == other.token_type


class Tokenizer(object):
    """
    Класс токенизации.
    """
    @staticmethod
    def tokenize(text: str, remove_punct=False, remove_unknown=False, replace_numbers=False) -> List[Token]:
        """
        Токенизация текстов на русском языке с учётом знаков препинания и слов с дефисами.

        :param text: исходный текст.
        :return: список токенов.
        """
        tokens = []
        punctuation = ".,?:;!—"
        begin = -1
        for i, ch in enumerate(text):
            if ch.isalpha() or ch == "-":
                if begin == -1:
                    begin = i
            else:
                if begin != -1:
                    tokens.append(Tokenizer.__form_token(text, begin, i))
                    begin = -1
                token_type = Token.TokenType.UNKNOWN
                if ch in punctuation:
                    token_type = Token.TokenType.PUNCTUATION
                elif ch == "\n":
                    token_type = Token.TokenType.ENDLINE
                elif ch == " ":
                    token_type = Token.TokenType.SPACE
                elif ch.isdigit():
                    token_type = Token.TokenType.NUMBER
                if len(tokens) != 0 and tokens[-1].token_type == token_type:
                    tokens[-1].text += ch
                    tokens[-1].end += 1
                else:
                    tokens.append(Token(ch, token_type, i, i + 1))
        if begin != -1:
            tokens.append(Tokenizer.__form_token(text, begin, len(text)))
        tokens = Tokenizer.__hyphen_map(tokens)
        if remove_punct:
            tokens = [token for token in tokens if token.token_type != Token.TokenType.PUNCTUATION]
        if remove_unknown:
            tokens = [token for token in tokens if token.token_type != Token.TokenType.UNKNOWN]
        if replace_numbers:
            for token in tokens:
                if token.token_type != Token.TokenType.NUMBER:
                    continue
                token.text = "ЧИСЛО"
                token.token_type = Token.TokenType.WORD
        return tokens

    @staticmethod
    def __form_token(text, begin, end):
        word = text[begin:end]
        if word != "-":
            return Token(word, Token.TokenType.WORD, begin, end)
        else:
            return Token("-", Token.TokenType.PUNCTUATION, begin, begin + 1)

    @staticmethod
    def __hyphen_map(tokens: List[Token]) -> List[Token]:
        """
        Слова из словаря оставляем с дефисом, остальные разделяем.

        :param tokens: токены.
        :return: токены после обработки.
        """
        new_tokens = []
        hyphen_tokens = Tokenizer.__get_hyphen_tokens()
        for token in tokens:
            if token.token_type != Token.TokenType.WORD:
                new_tokens.append(token)
                continue
            is_one_word = True
            if "-" in token.text:
                is_one_word = False
                for hyphen_token in hyphen_tokens:
                    if hyphen_token in token.text or token.text in hyphen_token:
                        is_one_word = True
            if is_one_word:
                new_tokens.append(token)
            else:
                texts = token.text.split("-")
                pos = token.begin
                for text in texts:
                    new_tokens.append(Token(text, Token.TokenType.WORD, pos, pos+len(text)))
                    pos += len(text) + 1
        return new_tokens

    @staticmethod
    def __get_hyphen_tokens():
        """
        :return: содержание словаря, в котором прописаны слова с дефисом.
        """
        with open(HYPHEN_TOKENS, "r", encoding="utf-8") as file:
            hyphen_tokens = [token.strip() for token in file.readlines()]
            return hyphen_tokens


class SentenceTokenizer(object):
    @staticmethod
    def tokenize(text: str) -> List[str]:
        m = re.split(r'(?<=[^А-ЯЁ].[^А-ЯЁ][.?!;]) +(?=[А-ЯЁ])', text)
        return m



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты для словаря ударений.

import unittest

from rupo.stress.dict import StressDict
from rupo.stress.word import Stress, StressedWord
from rupo.util.preprocess import VOWELS
from rupo.settings import RU_GRAPHEME_STRESS_PATH, ZALYZNYAK_DICT, RU_GRAPHEME_STRESS_TRIE_PATH


class TestStressDict(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.dict = StressDict(language="ru", zalyzniak_dict=ZALYZNYAK_DICT,
                              raw_dict_path=RU_GRAPHEME_STRESS_PATH, trie_path=RU_GRAPHEME_STRESS_TRIE_PATH)

    @classmethod
    def tearDownClass(cls):
        del cls.dict

    def test_get_stresses(self):
        self.assertCountEqual(self.dict.get_stresses("данный", Stress.Type.PRIMARY), [1])
        self.assertCountEqual(self.dict.get_stresses("союза", Stress.Type.PRIMARY), [2])
        self.assertCountEqual(self.dict.get_stresses("англосакс", Stress.Type.SECONDARY), [0])
        self.assertCountEqual(self.dict.get_stresses("англосакс", Stress.Type.ANY), [0, 6])
        self.assertCountEqual(self.dict.get_stresses("пора", Stress.Type.PRIMARY), [1, 3])

    def test_stress_only_in_vowels(self):
        for word, stresses in self.dict.get_all():
            for stress in stresses:
                self.assertIn(word[stress.position], VOWELS)




import os
from typing import List

from rupo.g2p.rnn import RNNG2PModel
from rupo.settings import RU_G2P_DEFAULT_MODEL, EN_G2P_DEFAULT_MODEL


class G2PPredictor:
    def predict(self, word: str) -> List[int]:
        raise NotImplementedError()


class RNNG2PPredictor:
    def __init__(self, language: str="ru", g2p_model_path: str=None):
        self.language = language
        self.g2p_model_path = g2p_model_path

        if language == "ru":
            self.__init_language_defaults(RU_G2P_DEFAULT_MODEL)
        elif language == "en":
            self.__init_language_defaults(EN_G2P_DEFAULT_MODEL)
        else:
            raise RuntimeError("Wrong language")

        if not os.path.exists(self.g2p_model_path):
            raise RuntimeError("No g2p model available (or wrong path)")

        self.g2p_model = RNNG2PModel(language=language)
        self.g2p_model.load(self.g2p_model_path)

    def __init_language_defaults(self, g2p_model_path):
        if self.g2p_model_path is None:
            self.g2p_model_path = g2p_model_path

    def predict(self, word: str) -> str:
        word = word.lower()
        return self.g2p_model.predict([word])[0]



# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Тесты марковских цепей.

import os
import unittest

from rupo.generate.language_model.markov import MarkovModelContainer
from rupo.main.vocabulary import StressVocabulary
from rupo.settings import EXAMPLES_DIR, MARKUP_XML_EXAMPLE


class TestMarkov(unittest.TestCase):
    def test_markov(self):
        for n in range(2, 5):
            vocab_dump_file = os.path.join(EXAMPLES_DIR, "vocab.pickle")
            markov_dump_file = os.path.join(EXAMPLES_DIR, "markov.pickle")
            vocabulary = StressVocabulary(vocab_dump_file, MARKUP_XML_EXAMPLE)
            markov = MarkovModelContainer(markov_dump_file, vocabulary, MARKUP_XML_EXAMPLE, n_grams=n)
            self.assertTrue(os.path.exists(vocab_dump_file))
            self.assertTrue(os.path.exists(markov_dump_file))
            os.remove(vocab_dump_file)
            os.remove(markov_dump_file)
            self.assertEqual(vocabulary.size()-n+1, len(markov.transitions))
            self.assertEqual(sum([sum(transition.values()) for transition in markov.transitions.values()]), vocabulary.size()-n+1)


# -*- coding: utf-8 -*-
# Автор: Гусев Илья
# Описание: Класс для удобной работы со словарём ударений.

import pygtrie
import os
import pickle
from typing import List, Dict, ItemsView, Set

from rupo.dict.cmu import CMUDict
from rupo.settings import RU_GRAPHEME_STRESS_PATH, RU_GRAPHEME_STRESS_TRIE_PATH, \
    EN_PHONEME_STRESS_PATH, EN_PHONEME_STRESS_TRIE_PATH, ZALYZNYAK_DICT, CMU_DICT

from rupo.stress.word import Stress


class StressDict:
    """
    Класс данных, для сериализации словаря как префиксного дерева и быстрой загрузки в память.
    """

    class Mode:
        GRAPHEMES = 0
        PHONEMES = 0

    def __init__(self, language: str="ru", mode: Mode=Mode.GRAPHEMES, raw_dict_path=None, trie_path=None,
                 zalyzniak_dict=ZALYZNYAK_DICT, cmu_dict=CMU_DICT) -> None:
        self.data = pygtrie.Trie()  # type: Dict[str, Set[Stress]]
        self.raw_dict_path = raw_dict_path
        self.trie_path = trie_path
        if language == "ru" and mode == self.Mode.GRAPHEMES:
            self.__init_defaults(RU_GRAPHEME_STRESS_PATH, RU_GRAPHEME_STRESS_TRIE_PATH)
            if not os.path.exists(self.raw_dict_path):
                from rupo.dict.zaliznyak import ZalyzniakDict
                ZalyzniakDict.convert_to_accent_only(zalyzniak_dict, self.raw_dict_path)
        elif mode == self.Mode.PHONEMES and language == "en":
            self.__init_defaults(EN_PHONEME_STRESS_PATH, EN_PHONEME_STRESS_TRIE_PATH)
            if not os.path.exists(self.raw_dict_path):
                CMUDict.convert_to_phoneme_stress(cmu_dict, self.raw_dict_path)
        else:
            assert False
        if not os.path.isfile(self.raw_dict_path):
            raise FileNotFoundError("Dictionary raw file not found.")
        if os.path.isfile(self.trie_path):
            self.load(self.trie_path)
        else:
            self.create(self.raw_dict_path, self.trie_path)

    def __init_defaults(self, raw_dict_path, trie_path):
        if self.raw_dict_path is None:
            self.raw_dict_path = raw_dict_path
        if self.trie_path is None:
            self.trie_path = trie_path

    def create(self, src_filename: str, dst_filename: str) -> None:
        """
        Загрузка словаря из файла.

        :param src_filename: имя файла с оригинальным словарём.
        :param dst_filename: имя файла, в который будет сохранён дамп.
        """
        with open(src_filename, 'r', encoding='utf-8') as f:
            for line in f:
                word, primary, secondary = line.split("\t")
                stresses = [Stress(int(a), Stress.Type.PRIMARY) for a in primary.strip().split(",")]
                if secondary.strip() != "":
                    stresses += [Stress(int(a), Stress.Type.SECONDARY) for a in secondary.strip().split(",")]
                self.update(word, stresses)
        self.save(dst_filename)

    def save(self, dst_filename: str) -> None:
        """
        Сохранение дампа.
        
        :param dst_filename: имя файла, в который сохраняем дамп словаря.
        """
        with open(dst_filename, "wb") as f:
            pickle.dump(self.data, f, pickle.HIGHEST_PROTOCOL)

    def load(self, dump_filename: str) -> None:
        """
        Загрузка дампа словаря.
        
        :param dump_filename: откуда загружаем.
        """
        with open(dump_filename, "rb") as f:
            self.data = pickle.load(f)

    def get_stresses(self, word: str, stress_type: Stress.Type=Stress.Type.ANY) -> List[int]:
        """
        Получение ударений нужного типа у слова.

        :param word: слово, которое мы хотим посмотреть в словаре.
        :param stress_type: тип ударения.
        :return forms: массив всех ударений.
        """
        if word in self.data:
            if stress_type == Stress.Type.ANY:
                return [stress.position for stress in self.data[word]]
            else:
                return [stress.position for stress in self.data[word] if stress.type == stress_type]
        return []

    def get_all(self) -> ItemsView[str, Set[Stress]]:
        """
        :return items: все ключи и ударения словаря.
        """
        return self.data.items()

    def update(self, word: str, stresses: List[Stress]) -> None:
        """
        Обновление словаря.

        :param word: слово.
        :param stresses: набор ударений.
        """
        if word not in self.data:
            self.data[word] = set(stresses)
        else:
            self.data[word].update(stresses)

    def update_primary_only(self, word: str, stresses: List[int]) -> None:
        self.update(word, [Stress(stress, Stress.Type.PRIMARY) for stress in stresses])


import os
from rupo.g2p.phonemes import Phonemes


class ZalyzniakDict:
    @staticmethod
    def convert_to_accent_only(dict_file, accent_file):
        with open(dict_file, 'r', encoding='utf-8') as r:
            lines = r.readlines()
        with open(accent_file, 'w', encoding='utf-8') as w:
            for line in lines:
                for word in line.split("#")[1].split(","):
                    word = word.strip()
                    pos = -1
                    clean_word = ""
                    primary = []
                    secondary = []
                    for i, ch in enumerate(word):
                        if ch == "'" or ch == "`":
                            if ch == "`":
                                secondary.append(pos)
                            else:
                                primary.append(pos)
                            continue
                        clean_word += ch
                        pos += 1
                        if ch == "ё":
                            primary.append(pos)
                    if len(primary) != 0:
                        w.write(clean_word + "\t" + ",".join([str(a) for a in primary]) + "\t" +
                                ",".join([str(a) for a in secondary]) + "\n")

    @staticmethod
    def convert_to_g2p_only(dict_file, g2p_dict_path, g2p_model):
        from rupo.g2p.rnn import RNNG2PModel
        g2p_predictor = RNNG2PModel()
        g2p_predictor.load(g2p_model)
        with open(dict_file, 'r', encoding='utf-8') as r:
            lines = r.readlines()
        with open(g2p_dict_path, 'w', encoding='utf-8') as w:
            words = []
            for line in lines:
                for word in line.split("#")[1].split(","):
                    word = word.strip()
                    clean_word = ""
                    for i, ch in enumerate(word):
                        if ch == "'" or ch == "`":
                            continue
                        clean_word += ch
                    words.append(clean_word)
            phonetic_words = g2p_predictor.predict(words)
            for i, word in enumerate(words):
                w.write(word + "\t" + phonetic_words[i] + "\n")

    @staticmethod
    def convert_to_phoneme_stress(source_file, destination_file, g2p_dict_path, g2p_model):
        from rupo.g2p.rnn import RNNG2PModel
        from rupo.g2p.aligner import Aligner
        from rupo.stress.dict import StressDict
        g2p_predictor = RNNG2PModel(g2p_dict_path)
        g2p_predictor.load(g2p_model)
        aligner = Aligner()
        grapheme_stress_dict_path = os.path.join(os.path.dirname(os.path.abspath(source_file)), "ru_grapheme_stress.txt")
        ZalyzniakDict.convert_to_accent_only(source_file, grapheme_stress_dict_path)
        d = StressDict(raw_dict_path=grapheme_stress_dict_path)
        vowels = set(Phonemes.VOWELS)
        with open(destination_file, 'w', encoding='utf-8') as w:
            samples = 0
            for word, accents in d.get_all():
                primary_in_dict = [int(stress[0]) for stress in accents if stress[1] == StressDict.StressType.PRIMARY]
                secondary_in_dict = [int(stress[0]) for stress in accents if stress[1] == StressDict.StressType.SECONDARY]
                phonemes = g2p_predictor.predict([word])[0]
                g, p = aligner.align(word, phonemes)
                primary = ZalyzniakDict.align_stresses(g, p, primary_in_dict)
                secondary = ZalyzniakDict.align_stresses(g, p, secondary_in_dict)
                is_valid = True
                for stress in primary+secondary:
                    if p[stress] not in vowels:
                        print(g, p, stress, p[stress])
                        is_valid = False
                if is_valid:
                    w.write(phonemes + "\t" + ",".join([str(i) for i in primary]) + "\t" +
                            ",".join([str(i) for i in secondary]) + "\n")
                samples += 1
                if samples % 1000 == 0:
                    print(samples)

    @staticmethod
    def align_stresses(g, p, stresses):
        index = -1
        spaces_count = 0
        for i, stress in enumerate(stresses):
            for j in range(len(p)):
                if g[j] == " ":
                    spaces_count += 1
                else:
                    index += 1
                    if index == stress:
                        stresses[i] += spaces_count
        return stresses


from rupo.main.markup import Markup, Line, Word, Syllable

MARKUP_EXAMPLE = Markup("Соломка король себя.\n Пора виться майкой в.", [
            Line(0, 20, "Соломка король себя.", [
                Word(0, 7, "Соломка",
                     [Syllable(0, 2, 0, "Со"),
                      Syllable(2, 5, 1, "лом", 3),
                      Syllable(5, 7, 2, "ка")]),
                Word(8, 14, "король",
                     [Syllable(0, 2, 0, "ко"),
                      Syllable(2, 6, 1, "роль", 3)]),
                Word(15, 19, "себя",
                     [Syllable(0, 2, 0, "се"),
                      Syllable(2, 4, 1, "бя", 3)])]),
            Line(21, 43, " Пора виться майкой в.",[
                Word(22, 26, "Пора",
                     [Syllable(0, 2, 0, "По", 1),
                      Syllable(2, 4, 1, "ра", 3)]),
                Word(27, 33, "виться",
                     [Syllable(0, 2, 0, "ви", 1),
                      Syllable(2, 6, 1, "ться")]),
                Word(34, 40, "майкой",
                     [Syllable(0, 3, 0, "май", 1),
                      Syllable(3, 6, 1, "кой")]),
                Word(41, 42, "в", [])
                ])])


from typing import List, Tuple

import numpy as np

from rupo.generate.prepare.grammeme_vectorizer import GrammemeVectorizer
from rupo.generate.prepare.word_form import WordForm
from rupo.generate.prepare.word_form_vocabulary import WordFormVocabulary, SEQ_END_WF
from rupo.util.tqdm_open import tqdm_open

CHAR_SET = " абвгдеёжзийклмнопрстуфхцчшщьыъэюяАБВГДЕЁЖЗИЙКЛМНОПРСТУФХЦЧШЩЬЫЪЭЮЯ" \
           "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ.,-'\""


class BatchGenerator:
    """
    Генератор наборов примеров для обучения.
    """
    def __init__(self, filenames: List[str], batch_size: int,
                 embedding_size: int, softmax_size: int, sentence_maxlen: int,
                 word_form_vocabulary: WordFormVocabulary, grammeme_vectorizer: GrammemeVectorizer,
                 max_word_len: int):
        """
        :param filenames: имена файлов с морфоразметкой.
        :param batch_size: размер набора семплов.
        :param softmax_size: размер выхода softmax-слоя (=размер итогового набора вероятностей)
        :param sentence_maxlen: маскимальная длина куска предложения.
        :param word_form_vocabulary: словарь словофрм.
        :param grammeme_vectorizer: векторизатор граммем.
        """
        self.filenames = filenames  # type: List[str]
        self.batch_size = batch_size  # type: int
        self.embedding_size = embedding_size # type: int
        self.softmax_size = softmax_size  # type: int
        self.sentence_maxlen = sentence_maxlen  # type: int
        self.max_word_len = max_word_len  # type: int
        self.word_form_vocabulary = word_form_vocabulary  # type: WordFormVocabulary
        self.grammeme_vectorizer = grammeme_vectorizer  # type: GrammemeVectorizer

    def __generate_seqs(self, sentences: List[List[WordForm]]) -> Tuple[List[List[WordForm]], List[WordForm]]:
        """
        Генерация семплов.
        
        :param sentences: куски предложений из словоформ.
        :return: пары (<семпл из словоформ>, следующая за ним словоформа (ответ)).
        """
        seqs, next_words = [], []
        for sentence in sentences:
            # Разворот для генерации справа налево.
            sentence = sentence[::-1]
            for i in range(1, len(sentence)):
                word_form = sentence[i]
                current_part = sentence[max(0, i-self.sentence_maxlen): i]
                # Если следующая словооформа не из предсказываемых, пропускаем её.
                if self.word_form_vocabulary.get_word_form_index(word_form) >= self.softmax_size:
                    continue
                seqs.append(current_part)
                next_words.append(word_form)
        return seqs, next_words

    def __to_tensor(self, sentences: List[List[WordForm]], next_words: List[WordForm]) -> \
            Tuple[np.array, np.array, np.array, np.array]:
        """
        Перевод семплов из словоформ в индексы словоформ, поиск грамматических векторов по индексу.
        
        :param sentences: семплы из словоформ.
        :param next_words: следующие за последовательностями из sentences слова.
        :return: индексы словоформ, грамматические векторы, ответы-индексы.
        """
        n_samples = len(sentences)

        lemmas = np.zeros((n_samples, self.sentence_maxlen), dtype=np.int)
        grammemes = np.zeros((n_samples, self.sentence_maxlen, self.grammeme_vectorizer.grammemes_count()), dtype=np.int)
        chars = np.zeros((n_samples, self.sentence_maxlen, self.max_word_len), dtype=np.int)
        y = np.zeros(n_samples, dtype=np.int)

        for i in range(n_samples):
            sentence = sentences[i]
            next_word = next_words[i]
            lemmas_vector, grammemes_vector, word_char_vectors = \
                self.get_sample(sentence, self.embedding_size, self.max_word_len,
                                word_form_vocabulary=self.word_form_vocabulary,
                                grammeme_vectorizer=self.grammeme_vectorizer)

            lemmas[i, -len(sentence):] = lemmas_vector
            grammemes[i, -len(sentence):] = grammemes_vector
            chars[i, -len(sentence):] = word_char_vectors
            y[i] = min(self.word_form_vocabulary.word_form_indices[next_word], self.softmax_size)
        return lemmas, grammemes, chars, y

    @staticmethod
    def get_sample(sentence, embedding_size: int, max_word_len: int, word_form_vocabulary: WordFormVocabulary,
                   grammeme_vectorizer: GrammemeVectorizer):
        lemmas_vector = [min(word_form_vocabulary.get_lemma_index(x), embedding_size) for x in sentence]
        grammemes_vector = [grammeme_vectorizer.get_vector_by_index(x.gram_vector_index) for x in sentence]
        word_char_vectors = []
        for word in sentence:
            char_indices = np.zeros(max_word_len)
            word_char_indices = [CHAR_SET.index(ch) if ch in CHAR_SET else len(CHAR_SET)
                                 for ch in word.text][:max_word_len]
            char_indices[-min(len(word.text), max_word_len):] = word_char_indices
            word_char_vectors.append(char_indices)
        return lemmas_vector, grammemes_vector, word_char_vectors

    def __iter__(self):
        """
        Получение очередного батча.
        
        :return: индексы словоформ, грамматические векторы, ответы-индексы.
        """
        for filename in self.filenames:
            sentences = [[]]
            next_words = []
            with tqdm_open(filename, encoding='utf-8') as f:
                for line in f:
                    line = line.strip()
                    word_form = self.__parse_line(line)
                    sentences[-1].append(word_form)
                    if word_form == SEQ_END_WF:
                        sentences.append([])
                    if len(sentences) >= self.batch_size:
                        sentences, next_words = self.__generate_seqs(sentences)
                        yield self.__to_tensor(sentences, next_words)
                        sentences = [[]]
            if len(sentences) != 0:
                yield self.__to_tensor(sentences, next_words)

    def __parse_line(self, line: str) -> WordForm:
        line = line.strip()
        try:
            if len(line) != 0:
                word, lemma, pos, tags = line.split('\t')[:4]
                word, lemma = word.lower(), lemma.lower() + '_' + pos
                gram_vector_index = self.grammeme_vectorizer.get_index_by_name(pos + "#" + tags)
                return WordForm(lemma, gram_vector_index, word)
        except ValueError:
            pass
        return SEQ_END_WF
